# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'

import io
import logging
import os
import threading
import time
import fnmatch
import datetime
import sys
import shutil
try:
	import queue
except ImportError:
	import Queue as queue
import requests

import octoprint.plugin
import octoprint.util as util

from octoprint.settings import settings
from octoprint.events import eventManager, Events
from octoprint.plugin import plugin_manager
from octoprint.util import monotonic_time
from octoprint.util import get_fully_qualified_classname as fqcn
from octoprint.util import sv
from octoprint.util.commandline import CommandlineCaller

import sarge
import collections

import re

try:
	from os import scandir
except ImportError:
	from scandir import scandir


# currently configured timelapse
current = None

# currently active render job, if any
current_render_job = None

# filename formats
_capture_format = "{prefix}-%d.jpg"
_capture_glob = "{prefix}-*.jpg"
_output_format = "{prefix}{postfix}.{extension}"

# ffmpeg progress regexes
_ffmpeg_duration_regex = re.compile(r"Duration: (\d{2}):(\d{2}):(\d{2})\.\d{2}")
_ffmpeg_current_regex = re.compile(r"time=(\d{2}):(\d{2}):(\d{2})\.\d{2}")

# old capture format, needed to delete old left-overs from
# versions <1.2.9
_old_capture_format_re = re.compile(r"^tmp_\d{5}.jpg$")

# valid timelapses
_valid_timelapse_types = ["off", "timed", "zchange"]

# callbacks for timelapse config updates
_update_callbacks = []

# lock for timelapse cleanup, must be re-entrant
_cleanup_lock = threading.RLock()

# lock for timelapse job
_job_lock = threading.RLock()

# cached valid timelapse extensions
_extensions = None

def valid_timelapse(path):
	global _extensions

	if _extensions is None:
		# create list of extensions
		extensions = ["mpg", "mpeg", "mp4", "m4v", "mkv"]

		hooks = plugin_manager().get_hooks("octoprint.timelapse.extensions")
		for name, hook in hooks.items():
			try:
				result = hook()
				if result is None or not isinstance(result, list):
					continue
				extensions += result
			except Exception:
				logging.getLogger(__name__).exception("Exception while retrieving additional timelapse "
				                                      "extensions from hook {name}".format(name=name),
				                                      extra=dict(plugin=name))

		_extensions = list(set(extensions))

	return util.is_allowed_file(path, _extensions)


def _extract_prefix(filename):
	"""
	>>> _extract_prefix("some_long_filename_without_hyphen.jpg")
	>>> _extract_prefix("-first_char_is_hyphen.jpg")
	>>> _extract_prefix("some_long_filename_with-stuff.jpg") # doctest: +ALLOW_UNICODE
	'some_long_filename_with'
	"""
	pos = filename.rfind("-")
	if not pos or pos < 0:
		return None
	return filename[:pos]


def last_modified_finished():
	return os.stat(settings().getBaseFolder("timelapse", check_writable=False)).st_mtime


def last_modified_unrendered():
	return os.stat(settings().getBaseFolder("timelapse_tmp", check_writable=False)).st_mtime


def get_finished_timelapses():
	files = []
	basedir = settings().getBaseFolder("timelapse", check_writable=False)
	for entry in scandir(basedir):
		if util.is_hidden_path(entry.path) or not valid_timelapse(entry.path):
			continue
		files.append({
			"name": entry.name,
			"size": util.get_formatted_size(entry.stat().st_size),
			"bytes": entry.stat().st_size,
			"date": util.get_formatted_datetime(datetime.datetime.fromtimestamp(entry.stat().st_mtime))
		})
	return files


def get_unrendered_timelapses():
	global _job_lock
	global current

	delete_old_unrendered_timelapses()

	basedir = settings().getBaseFolder("timelapse_tmp", check_writable=False)
	jobs = collections.defaultdict(lambda: dict(count=0, size=None, bytes=0, date=None, timestamp=None))

	for entry in scandir(basedir):
		if not fnmatch.fnmatch(entry.name, "*.jpg"):
			continue

		prefix = _extract_prefix(entry.name)
		if prefix is None:
			continue

		jobs[prefix]["count"] += 1
		jobs[prefix]["bytes"] += entry.stat().st_size
		if jobs[prefix]["timestamp"] is None or entry.stat().st_mtime < jobs[prefix]["timestamp"]:
			jobs[prefix]["timestamp"] = entry.stat().st_mtime

	with _job_lock:
		global current_render_job

		def finalize_fields(prefix, job):
			currently_recording = current is not None and current.prefix == prefix
			currently_rendering = current_render_job is not None and current_render_job["prefix"] == prefix

			job["size"] = util.get_formatted_size(job["bytes"])
			job["date"] = util.get_formatted_datetime(datetime.datetime.fromtimestamp(job["timestamp"]))
			job["recording"] = currently_recording
			job["rendering"] = currently_rendering
			job["processing"] = currently_recording or currently_rendering
			del job["timestamp"]

			return job

		return sorted([util.dict_merge(dict(name=key), finalize_fields(key, value))
		               for key, value in jobs.items()],
		              key=lambda x: sv(x["name"]))


def delete_unrendered_timelapse(name):
	global _cleanup_lock

	pattern = "{}*.jpg".format(util.glob_escape(name))

	basedir = settings().getBaseFolder("timelapse_tmp")
	with _cleanup_lock:
		for entry in scandir(basedir):
			try:
				if fnmatch.fnmatch(entry.name, pattern):
					os.remove(entry.path)
			except Exception:
				if logging.getLogger(__name__).isEnabledFor(logging.DEBUG):
					logging.getLogger(__name__).exception("Error while processing file {} during cleanup".format(entry.name))


def render_unrendered_timelapse(name, gcode=None, postfix=None, fps=None):
	capture_dir = settings().getBaseFolder("timelapse_tmp")
	output_dir = settings().getBaseFolder("timelapse")

	if fps is None:
		fps = settings().getInt(["webcam", "timelapse", "fps"])
	threads = settings().get(["webcam", "ffmpegThreads"])
	videocodec = settings().get(["webcam", "ffmpegVideoCodec"])

	job = TimelapseRenderJob(capture_dir, output_dir, name,
	                         postfix=postfix,
	                         fps=fps,
	                         threads=threads,
	                         videocodec=videocodec,
	                         on_start=_create_render_start_handler(name, gcode=gcode),
	                         on_success=_create_render_success_handler(name, gcode=gcode),
	                         on_fail=_create_render_fail_handler(name, gcode=gcode),
	                         on_always=_create_render_always_handler(name, gcode=gcode))
	job.process()


def delete_old_unrendered_timelapses():
	global _cleanup_lock

	basedir = settings().getBaseFolder("timelapse_tmp")
	clean_after_days = settings().getInt(["webcam", "cleanTmpAfterDays"])
	cutoff = time.time() - clean_after_days * 24 * 60 * 60

	prefixes_to_clean = []

	with _cleanup_lock:
		for entry in scandir(basedir):
			try:
				prefix = _extract_prefix(entry.name)
				if prefix is None:
					# might be an old tmp_00000.jpg kinda frame. we can't
					# render those easily anymore, so delete that stuff
					if _old_capture_format_re.match(entry.name):
						os.remove(entry.path)
					continue

				if prefix in prefixes_to_clean:
					continue

				# delete if both creation and modification time are older than the cutoff
				if max(entry.stat().st_ctime, entry.stat().st_mtime) < cutoff:
					prefixes_to_clean.append(prefix)
			except Exception:
				if logging.getLogger(__name__).isEnabledFor(logging.DEBUG):
					logging.getLogger(__name__).exception("Error while processing file {} during cleanup".format(entry.name))

		for prefix in prefixes_to_clean:
			delete_unrendered_timelapse(prefix)
			logging.getLogger(__name__).info("Deleted old unrendered timelapse {}".format(prefix))


def _create_render_start_handler(name, gcode=None):
	def f(movie):
		global _job_lock

		with _job_lock:
			global current_render_job
			payload = dict(gcode=gcode if gcode is not None else "unknown",
			               movie=movie,
			               movie_basename=os.path.basename(movie),
			               movie_prefix=name)
			current_render_job = dict(prefix=name)
			current_render_job.update(payload)
		eventManager().fire(Events.MOVIE_RENDERING, payload)
	return f


def _create_render_success_handler(name, gcode=None):
	def f(movie):
		delete_unrendered_timelapse(name)
		payload = dict(gcode=gcode if gcode is not None else "unknown",
		               movie=movie,
		               movie_basename=os.path.basename(movie),
		               movie_prefix=name)
		eventManager().fire(Events.MOVIE_DONE, payload)
	return f


def _create_render_fail_handler(name, gcode=None):
	def f(movie, returncode=255, stdout="Unknown error", stderr="Unknown error", reason="unknown"):
		payload = dict(gcode=gcode if gcode is not None else "unknown",
		               movie=movie,
		               movie_basename=os.path.basename(movie),
		               movie_prefix=name,
		               returncode=returncode,
		               out=stdout,
		               error=stderr,
		               reason=reason)
		eventManager().fire(Events.MOVIE_FAILED, payload)
	return f


def _create_render_always_handler(name, gcode=None):
	def f(movie):
		global current_render_job
		global _job_lock
		with _job_lock:
			current_render_job = None
	return f


def register_callback(callback):
	if not callback in _update_callbacks:
		_update_callbacks.append(callback)


def unregister_callback(callback):
	try:
		_update_callbacks.remove(callback)
	except ValueError:
		# not registered
		pass


def notify_callbacks(timelapse):
	if timelapse is None:
		config = None
	else:
		config = timelapse.config_data()

	for callback in _update_callbacks:
		notify_callback(callback, config)


def notify_callback(callback, config=None, timelapse=None):
	if config is None and timelapse is not None:
			config = timelapse.config_data()

	try:
		callback.sendTimelapseConfig(config)
	except Exception:
		logging.getLogger(__name__).exception("Exception while pushing timelapse configuration",
		                                      extra=dict(callback=fqcn(callback)))


def configure_timelapse(config=None, persist=False):
	global current

	if config is None:
		config = settings().get(["webcam", "timelapse"], merged=True)

	if current is not None:
		current.unload()

	snapshot_url = settings().get(["webcam", "snapshot"])
	ffmpeg_path = settings().get(["webcam", "ffmpeg"])
	timelapse_enabled = settings().getBoolean(["webcam", "timelapseEnabled"])
	timelapse_precondition = snapshot_url is not None and snapshot_url.strip() != "" \
	                         and ffmpeg_path is not None and ffmpeg_path.strip() != ""

	type = config["type"]
	if not timelapse_precondition and timelapse_precondition:
		logging.getLogger(__name__).warn("Essential timelapse settings unconfigured (snapshot URL or FFMPEG path) "
		                                 "but timelapse enabled.")

	if not timelapse_enabled or not timelapse_precondition or type is None or "off" == type:
		current = None

	else:
		postRoll = 0
		if "postRoll" in config and config["postRoll"] >= 0:
			postRoll = config["postRoll"]

		fps = 25
		if "fps" in config and config["fps"] > 0:
			fps = config["fps"]

		if "zchange" == type:
			retractionZHop = 0
			if "options" in config and "retractionZHop" in config["options"] and config["options"]["retractionZHop"] >= 0:
				retractionZHop = config["options"]["retractionZHop"]

			minDelay = 5
			if "options" in config and "minDelay" in config["options"] and config["options"]["minDelay"] > 0:
				minDelay = config["options"]["minDelay"]

			current = ZTimelapse(post_roll=postRoll, retraction_zhop=retractionZHop, min_delay=minDelay, fps=fps)

		elif "timed" == type:
			interval = 10
			if "options" in config and "interval" in config["options"] and config["options"]["interval"] > 0:
				interval = config["options"]["interval"]

			current = TimedTimelapse(post_roll=postRoll, interval=interval, fps=fps)

	notify_callbacks(current)

	if persist:
		settings().set(["webcam", "timelapse"], config)
		settings().save()


class Timelapse(object):
	QUEUE_ENTRY_TYPE_CAPTURE = "capture"
	QUEUE_ENTRY_TYPE_CALLBACK = "callback"

	def __init__(self, post_roll=0, fps=25):
		self._logger = logging.getLogger(__name__)
		self._image_number = None
		self._in_timelapse = False
		self._gcode_file = None
		self._file_prefix = None

		self._capture_errors = 0
		self._capture_success = 0

		self._post_roll = post_roll
		self._on_post_roll_done = None

		self._capture_dir = settings().getBaseFolder("timelapse_tmp")
		self._movie_dir = settings().getBaseFolder("timelapse")
		self._snapshot_url = settings().get(["webcam", "snapshot"])
		self._snapshot_timeout = settings().getInt(["webcam", "snapshotTimeout"])
		self._snapshot_validate_ssl = settings().getBoolean(["webcam", "snapshotSslValidation"])

		self._fps = fps


		self._pluginManager = octoprint.plugin.plugin_manager()
		self._pre_capture_hooks = self._pluginManager.get_hooks("octoprint.timelapse.capture.pre")
		self._post_capture_hooks = self._pluginManager.get_hooks("octoprint.timelapse.capture.post")

		self._capture_mutex = threading.Lock()
		self._capture_queue = queue.Queue()
		self._capture_queue_active = True

		self._capture_queue_thread = threading.Thread(target=self._capture_queue_worker)
		self._capture_queue_thread.daemon = True
		self._capture_queue_thread.start()

		# subscribe events
		eventManager().subscribe(Events.PRINT_STARTED, self.on_print_started)
		eventManager().subscribe(Events.PRINT_FAILED, self.on_print_done)
		eventManager().subscribe(Events.PRINT_DONE, self.on_print_done)
		eventManager().subscribe(Events.PRINT_RESUMED, self.on_print_resumed)
		for (event, callback) in self.event_subscriptions():
			eventManager().subscribe(event, callback)

	@property
	def prefix(self):
		return self._file_prefix

	@property
	def post_roll(self):
		return self._post_roll

	@property
	def fps(self):
		return self._fps

	def unload(self):
		if self._in_timelapse:
			self.stop_timelapse(do_create_movie=False)

		# unsubscribe events
		eventManager().unsubscribe(Events.PRINT_STARTED, self.on_print_started)
		eventManager().unsubscribe(Events.PRINT_FAILED, self.on_print_done)
		eventManager().unsubscribe(Events.PRINT_DONE, self.on_print_done)
		eventManager().unsubscribe(Events.PRINT_RESUMED, self.on_print_resumed)
		for (event, callback) in self.event_subscriptions():
			eventManager().unsubscribe(event, callback)

	def on_print_started(self, event, payload):
		"""
		Override this to perform additional actions upon start of a print job.
		"""
		self.start_timelapse(payload["name"])

	def on_print_done(self, event, payload):
		"""
		Override this to perform additional actions upon the stop of a print job.
		"""
		self.stop_timelapse(success=(event==Events.PRINT_DONE))

	def on_print_resumed(self, event, payload):
		"""
		Override this to perform additional actions upon the pausing of a print job.
		"""
		if not self._in_timelapse:
			self.start_timelapse(payload["name"])

	def event_subscriptions(self):
		"""
		Override this method to subscribe to additional events by returning an array of (event, callback) tuples.

		Events that are already subscribed:
		  * PrintStarted - self.onPrintStarted
		  * PrintResumed - self.onPrintResumed
		  * PrintFailed - self.onPrintDone
		  * PrintDone - self.onPrintDone
		"""
		return []

	def config_data(self):
		"""
		Override this method to return the current timelapse configuration data. The data should have the following
		form:

		    type: "<type of timelapse>",
		    options: { <additional options> }
		"""
		return None

	def start_timelapse(self, gcode_file):
		self._logger.debug("Starting timelapse for %s" % gcode_file)

		self._image_number = 0
		self._capture_errors = 0
		self._capture_success = 0
		self._in_timelapse = True
		self._gcode_file = os.path.basename(gcode_file)
		self._file_prefix = "{}_{}".format(os.path.splitext(self._gcode_file)[0], time.strftime("%Y%m%d%H%M%S"))

	def stop_timelapse(self, do_create_movie=True, success=True):
		self._logger.debug("Stopping timelapse")

		self._in_timelapse = False

		def reset_image_number():
			self._image_number = None

		def create_movie():
			render_unrendered_timelapse(self._file_prefix,
			                            gcode=self._gcode_file,
			                            postfix=None if success else "-fail",
			                            fps=self._fps)

		def reset_and_create():
			reset_image_number()
			create_movie()

		def wait_for_captures(callback):
			self._capture_queue.put(dict(type=self.__class__.QUEUE_ENTRY_TYPE_CALLBACK, callback=callback))

		def create_wait_for_captures(callback):
			def f():
				wait_for_captures(callback)
			return f

		# wait for everything so far in the queue to be processed, then see if we should process from there
		def continue_rendering():
			if self._capture_success == 0:
				# no images - either nothing was attempted to be captured or all attempts ran into an error
				if self._capture_errors > 0:
					# this is the latter case
					_create_render_fail_handler(self._file_prefix,
					                            gcode=self._gcode_file)("n/a",
					                                                    returncode=0,
					                                                    stdout="",
					                                                    stderr="",
					                                                    reason="no_frames")

				# in any case, don't continue
				return

			# check if we have post roll configured
			if self._post_roll > 0:
				# capture post roll, wait for THAT to finish, THEN render
				eventManager().fire(Events.POSTROLL_START,
				                    dict(postroll_duration=self.calculate_post_roll(),
				                         postroll_length=self.post_roll,
				                         postroll_fps=self.fps))
				if do_create_movie:
					self._on_post_roll_done = create_wait_for_captures(reset_and_create)
				else:
					self._on_post_roll_done = reset_image_number
				self.process_post_roll()
			else:
				# no post roll? perfect, render
				if do_create_movie:
					wait_for_captures(reset_and_create)
				else:
					reset_image_number()

		self._logger.debug("Waiting to process capture queue")
		wait_for_captures(continue_rendering)

	def calculate_post_roll(self):
		return None

	def process_post_roll(self):
		self.post_roll_finished()

	def post_roll_finished(self):
		if self.post_roll:
			eventManager().fire(Events.POSTROLL_END)
			if self._on_post_roll_done is not None:
				self._on_post_roll_done()

	def capture_image(self):
		if self._capture_dir is None:
			self._logger.warning("Cannot capture image, capture directory is unset")
			return

		with self._capture_mutex:
			if self._image_number is None:
				self._logger.warning("Cannot capture image, image number is unset")
				return

			filename = os.path.join(self._capture_dir, _capture_format.format(prefix=self._file_prefix) % self._image_number)
			self._image_number += 1

		self._logger.debug("Capturing image to {}".format(filename))
		entry = dict(type=self.__class__.QUEUE_ENTRY_TYPE_CAPTURE,
		             filename=filename,
		             onerror=self._on_capture_error)
		self._capture_queue.put(entry)
		return filename

	def _on_capture_error(self):
		with self._capture_mutex:
			if self._image_number is not None and self._image_number > 0:
				self._image_number -= 1

	def _capture_queue_worker(self):
		while self._capture_queue_active:
			entry = self._capture_queue.get(block=True)

			if entry["type"] == self.__class__.QUEUE_ENTRY_TYPE_CAPTURE and "filename" in entry:
				filename = entry["filename"]
				onerror = entry.pop("onerror", None)
				self._perform_capture(filename, onerror=onerror)

			elif entry["type"] == self.__class__.QUEUE_ENTRY_TYPE_CALLBACK and "callback" in entry:
				args = entry.pop("args", [])
				kwargs = entry.pop("kwargs", dict())
				entry["callback"](*args, **kwargs)

	def _perform_capture(self, filename, onerror=None):
		# pre-capture hook
		for name, hook in self._pre_capture_hooks.items():
			try:
				hook(filename)
			except Exception:
				self._logger.exception("Error while processing hook {name}.".format(**locals()))

		eventManager().fire(Events.CAPTURE_START, dict(file=filename))
		try:
			self._logger.debug("Going to capture {} from {}".format(filename, self._snapshot_url))
			r = requests.get(self._snapshot_url,
			                 stream=True,
			                 timeout=self._snapshot_timeout,
			                 verify=self._snapshot_validate_ssl)
			r.raise_for_status()

			with io.open(filename, 'wb') as f:
				for chunk in r.iter_content(chunk_size=1024):
					if chunk:
						f.write(chunk)
						f.flush()

			self._logger.debug("Image {} captured from {}".format(filename, self._snapshot_url))
		except Exception as e:
			self._logger.exception("Could not capture image {} from {}".format(filename, self._snapshot_url))
			self._capture_errors += 1
			err = e
		else:
			self._capture_success += 1
			err = None

		# post-capture hook
		for name, hook in self._post_capture_hooks.items():
			try:
				hook(filename, err is None)
			except Exception:
				self._logger.exception("Error while processing hook {name}.".format(**locals()))

		# handle events and onerror call
		if err is None:
			eventManager().fire(Events.CAPTURE_DONE, dict(file=filename))
			return True
		else:
			if callable(onerror):
				onerror()
			eventManager().fire(Events.CAPTURE_FAILED, dict(file=filename,
			                                                error=str(err),
			                                                url=self._snapshot_url))
			return False


	def _copying_postroll(self):
		with self._capture_mutex:
			filename = os.path.join(self._capture_dir,
			                        _capture_format.format(prefix=self._file_prefix) % self._image_number)
			self._image_number += 1

		if self._perform_capture(filename):
			for _ in range(self._post_roll * self._fps):
				newFile = os.path.join(self._capture_dir,
				                       _capture_format.format(prefix=self._file_prefix) % self._image_number)
				self._image_number += 1
				shutil.copyfile(filename, newFile)

	def clean_capture_dir(self):
		if not os.path.isdir(self._capture_dir):
			self._logger.warning("Cannot clean capture directory, it is unset")
			return
		delete_unrendered_timelapse(self._file_prefix)



class ZTimelapse(Timelapse):
	def __init__(self, retraction_zhop=0, min_delay=5.0, post_roll=0, fps=25):
		Timelapse.__init__(self, post_roll=post_roll, fps=fps)

		if min_delay < 0:
			min_delay = 0

		self._retraction_zhop = retraction_zhop
		self._min_delay = min_delay
		self._last_snapshot = None
		self._logger.debug("ZTimelapse initialized")

	@property
	def retraction_zhop(self):
		return self._retraction_zhop

	@property
	def min_delay(self):
		return self._min_delay

	def event_subscriptions(self):
		return [
			(Events.Z_CHANGE, self._on_z_change)
		]

	def config_data(self):
		return {
			"type": "zchange",
			"options": {
				"retractionZHop": self._retraction_zhop
			}
		}

	def process_post_roll(self):
		# we always copy the final image for the whole post roll
		# for z based timelapses
		self._copying_postroll()
		Timelapse.process_post_roll(self)

	def _on_z_change(self, event, payload):
		# check if height difference equals z-hop, if so don't take a picture
		if self._retraction_zhop != 0 and payload["old"] is not None and payload["new"] is not None:
			diff = round(abs(payload["new"] - payload["old"]), 3)
			zhop = round(self._retraction_zhop, 3)
			if diff == zhop:
				return

		# check if last picture has been less than min_delay ago, if so don't take a picture (anti vase mode...)
		now = monotonic_time()
		if self._min_delay and self._last_snapshot and self._last_snapshot + self._min_delay > now:
			self._logger.debug("Rate limited z-change, not taking a snapshot")
			return

		self.capture_image()
		self._last_snapshot = now


class TimedTimelapse(Timelapse):
	def __init__(self, interval=1, post_roll=0, fps=25):
		Timelapse.__init__(self, post_roll=post_roll, fps=fps)
		self._interval = interval
		if self._interval < 1:
			self._interval = 1 # force minimum interval of 1s
		self._timer = None
		self._logger.debug("TimedTimelapse initialized")

	@property
	def interval(self):
		return self._interval

	def config_data(self):
		return {
			"type": "timed",
			"options": {
				"interval": self._interval
			}
		}

	def on_print_started(self, event, payload):
		Timelapse.on_print_started(self, event, payload)
		if self._timer is not None:
			return

		self._logger.debug("Starting timer for interval based timelapse")
		from octoprint.util import RepeatedTimer
		self._timer = RepeatedTimer(self._interval, self._timer_task,
		                            run_first=True, condition=self._timer_active,
		                            on_finish=self._on_timer_finished)
		self._timer.start()

	def process_post_roll(self):
		# we only use the final image as post roll
		self._copying_postroll()
		self.post_roll_finished()

	def _timer_active(self):
		return self._in_timelapse

	def _timer_task(self):
		self.capture_image()

	def _on_timer_finished(self):
		# timer is done, delete it
		self._timer = None


class TimelapseRenderJob(object):

	render_job_lock = threading.RLock()

	def __init__(self, capture_dir, output_dir, prefix, postfix=None, capture_glob=_capture_glob,
	             capture_format=_capture_format, output_format=_output_format, fps=25, threads=1,
	             videocodec="mpeg2video", on_start=None, on_success=None, on_fail=None, on_always=None):
		self._capture_dir = capture_dir
		self._output_dir = output_dir
		self._prefix = prefix
		self._postfix = postfix
		self._capture_glob = capture_glob
		self._capture_format = capture_format
		self._output_format = output_format
		self._fps = fps
		self._threads = threads
		self._videocodec = videocodec
		self._on_start = on_start
		self._on_success = on_success
		self._on_fail = on_fail
		self._on_always = on_always

		self._thread = None
		self._logger = logging.getLogger(__name__)

		self._parsed_duration = 0

	def process(self):
		"""Processes the job."""

		self._thread = threading.Thread(target=self._render,
		                                name="TimelapseRenderJob_{prefix}_{postfix}".format(prefix=self._prefix,
		                                                                                    postfix=self._postfix))
		self._thread.daemon = True
		self._thread.start()

	def _render(self):
		"""Rendering runnable."""

		ffmpeg = settings().get(["webcam", "ffmpeg"])
		bitrate = settings().get(["webcam", "bitrate"])
		if ffmpeg is None or bitrate is None:
			self._logger.warning("Cannot create movie, path to ffmpeg or desired bitrate is unset")
			return

		if self._videocodec == 'mpeg2video':
			extension = "mpg"
		else:
			extension = "mp4"

		input = os.path.join(self._capture_dir,
		                     self._capture_format.format(prefix=self._prefix,
		                                                 postfix=self._postfix if self._postfix is not None else ""))

		output_name = self._output_format.format(prefix=self._prefix,
		                                         postfix=self._postfix if self._postfix is not None else "",
		                                         extension=extension)
		temporary = os.path.join(self._output_dir, ".{}".format(output_name))
		output = os.path.join(self._output_dir, output_name)

		for i in range(4):
			if os.path.exists(input % i):
				break
		else:
			self._logger.warning("Cannot create a movie, no frames captured")
			self._notify_callback("fail", output, returncode=0, stdout="", stderr="", reason="no_frames")
			return

		hflip = settings().getBoolean(["webcam", "flipH"])
		vflip = settings().getBoolean(["webcam", "flipV"])
		rotate = settings().getBoolean(["webcam", "rotate90"])

		watermark = None
		if settings().getBoolean(["webcam", "watermark"]):
			watermark = os.path.join(os.path.dirname(__file__), "static", "img", "watermark.png")
			if sys.platform == "win32":
				# Because ffmpeg hiccups on windows' drive letters and backslashes we have to give the watermark
				# path a special treatment. Yeah, I couldn't believe it either...
				watermark = watermark.replace("\\", "/").replace(":", "\\\\:")

		# prepare ffmpeg command
		command_str = self._create_ffmpeg_command_string(ffmpeg, self._fps, bitrate, self._threads, input, temporary,
		                                                 self._videocodec, hflip=hflip, vflip=vflip, rotate=rotate,
		                                                 watermark=watermark)
		self._logger.debug("Executing command: {}".format(command_str))

		with self.render_job_lock:
			try:
				self._notify_callback("start", output)

				self._logger.debug("Parsing ffmpeg output")

				c = CommandlineCaller()
				c.on_log_stderr = self._process_ffmpeg_output
				returncode, stdout_text, stderr_text = c.call(command_str, delimiter=b'\r', buffer_size=512)

				self._logger.debug("Done with parsing")

				if returncode == 0:
					shutil.move(temporary, output)
					self._notify_callback("success", output)
				else:
					self._logger.warning("Could not render movie, got return code %r: %s" % (returncode, stderr_text))
					self._notify_callback("fail", output, returncode=returncode, stdout=stdout_text, stderr=stderr_text, reason="returncode")
			except Exception:
				self._logger.exception("Could not render movie due to unknown error")
				self._notify_callback("fail", output, reason="unknown")
			finally:
				try:
					if os.path.exists(temporary):
						os.remove(temporary)
				except Exception:
					self._logger.warn("Could not delete temporary timelapse {}".format(temporary))
				self._notify_callback("always", output)

	def _process_ffmpeg_output(self, *lines):
		for line in lines:
			# We should be getting the time more often, so try it first
			current_time = _ffmpeg_current_regex.search(line)
			if current_time is not None and self._parsed_duration != 0:
				current_s = self._convert_time(*current_time.groups())
				progress = current_s / float(self._parsed_duration) * 100

				# Update progress bar
				for callback in _update_callbacks:
					try:
						callback.sendRenderProgress(progress)
					except Exception:
						self._logger.exception("Exception while pushing render progress")

			else:
				duration = _ffmpeg_duration_regex.search(line)
				if duration is not None:
					self._parsed_duration = self._convert_time(*duration.groups())

	@staticmethod
	def _convert_time(hours, minutes, seconds):
		return (int(hours) * 60 + int(minutes)) * 60 + int(seconds)

	@classmethod
	def _create_ffmpeg_command_string(cls, ffmpeg, fps, bitrate, threads, input, output, videocodec, hflip=False, vflip=False,
	                                  rotate=False, watermark=None, pixfmt="yuv420p"):
		"""
		Create ffmpeg command string based on input parameters.

		Arguments:
		    ffmpeg (str): Path to ffmpeg
		    fps (int): Frames per second for output
		    bitrate (str): Bitrate of output
		    threads (int): Number of threads to use for rendering
		    videocodec (str): Videocodec to be used for encoding
		    input (str): Absolute path to input files including file mask
		    output (str): Absolute path to output file
		    hflip (bool): Perform horizontal flip on input material.
		    vflip (bool): Perform vertical flip on input material.
		    rotate (bool): Perform 90° CCW rotation on input material.
		    watermark (str): Path to watermark to apply to lower left corner.
		    pixfmt (str): Pixel format to use for output. Default of yuv420p should usually fit the bill.

		Returns:
		    (str): Prepared command string to render `input` to `output` using ffmpeg.
		"""

		### See unit tests in test/timelapse/test_timelapse_renderjob.py

		logger = logging.getLogger(__name__)

		### Not all players can handle non-mpeg2 in VOB format
		if videocodec == "mpeg2video":
			containerformat = "vob"
		else:
			containerformat = "mp4"

		command = [
			ffmpeg, '-framerate', str(fps), '-i', '"{}"'.format(input), '-vcodec', videocodec,
			'-threads', str(threads), '-r', "25", '-y', '-b', str(bitrate),
			'-f', containerformat]

		filter_string = cls._create_filter_string(hflip=hflip,
		                                          vflip=vflip,
		                                          rotate=rotate,
		                                          watermark=watermark)

		if filter_string is not None:
			logger.debug("Applying videofilter chain: {}".format(filter_string))
			command.extend(["-vf", sarge.shell_quote(filter_string)])

		# finalize command with output file
		logger.debug("Rendering movie to {}".format(output))
		command.append('"{}"'.format(output))

		return " ".join(command)

	@classmethod
	def _create_filter_string(cls, hflip=False, vflip=False, rotate=False, watermark=None, pixfmt="yuv420p"):
		"""
		Creates an ffmpeg filter string based on input parameters.

		Arguments:
		    hflip (bool): Perform horizontal flip on input material.
		    vflip (bool): Perform vertical flip on input material.
		    rotate (bool): Perform 90° CCW rotation on input material.
		    watermark (str): Path to watermark to apply to lower left corner.
		    pixfmt (str): Pixel format to use, defaults to "yuv420p" which should usually fit the bill

		Returns:
		    (str or None): filter string or None if no filters are required
		"""

		### See unit tests in test/timelapse/test_timelapse_renderjob.py

		# apply pixel format
		filters = ["format={}".format(pixfmt)]

		# flip video if configured
		if hflip:
			filters.append('hflip')
		if vflip:
			filters.append('vflip')
		if rotate:
			filters.append('transpose=2')

		# add watermark if configured
		watermark_filter = None
		if watermark is not None:
			watermark_filter = "movie={} [wm]; [{{input_name}}][wm] overlay=10:main_h-overlay_h-10".format(watermark)

		filter_string = None
		if len(filters) > 0:
			if watermark_filter is not None:
				filter_string = "[in] {} [postprocessed]; {} [out]".format(",".join(filters),
				                                                           watermark_filter.format(input_name="postprocessed"))
			else:
				filter_string = "[in] {} [out]".format(",".join(filters))
		elif watermark_filter is not None:
			filter_string = watermark_filter.format(input_name="in") + " [out]"

		return filter_string

	def _notify_callback(self, callback, *args, **kwargs):
		"""Notifies registered callbacks of type `callback`."""
		name = "_on_{}".format(callback)
		method = getattr(self, name, None)
		if method is not None and callable(method):
			method(*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


import logging
try:
	import queue
except ImportError:
	import Queue as queue
import os
import threading
import collections
import time

from octoprint.events import Events, eventManager
from octoprint.settings import settings
from octoprint.util import monotonic_time
from octoprint.util import get_fully_qualified_classname as fqcn


class QueueEntry(collections.namedtuple("QueueEntry", "name, path, type, location, absolute_path, printer_profile, analysis")):
	"""
	A :class:`QueueEntry` for processing through the :class:`AnalysisQueue`. Wraps the entry's properties necessary
	for processing.

	Arguments:
	    name (str): Name of the file to analyze.
	    path (str): Storage location specific path to the file to analyze.
	    type (str): Type of file to analyze, necessary to map to the correct :class:`AbstractAnalysisQueue` sub class.
	        At the moment, only ``gcode`` is supported here.
	    location (str): Location the file is located on.
	    absolute_path (str): Absolute path on disk through which to access the file.
	    printer_profile (PrinterProfile): :class:`PrinterProfile` which to use for analysis.
	    analysis (dict): :class:`GcodeAnalysisQueue` results from prior analysis, or ``None`` if there is none.
	"""

	def __str__(self):
		return "{location}:{path}".format(location=self.location, path=self.path)


class AnalysisAborted(Exception):
	def __init__(self, reenqueue=True, *args, **kwargs):
		Exception.__init__(self, *args, **kwargs)
		self.reenqueue = reenqueue


class AnalysisQueue(object):
	"""
	OctoPrint's :class:`AnalysisQueue` can manage various :class:`AbstractAnalysisQueue` implementations, mapped
	by their machine code type.

	By invoking :meth:`register_finish_callback` it is possible to register oneself as a callback to be invoked each
	time the analysis of a queue entry finishes. The call parameters will be the finished queue entry as the first
	and the analysis result as the second parameter. It is also possible to remove the registration again by invoking
	:meth:`unregister_finish_callback`.

	:meth:`enqueue` allows enqueuing :class:`QueueEntry` instances to analyze. If the :attr:`QueueEntry.type` is unknown
	(no specific child class of :class:`AbstractAnalysisQueue` is registered for it), nothing will happen. Otherwise the
	entry will be enqueued with the type specific analysis queue.
	"""

	def __init__(self, queue_factories):
		self._logger = logging.getLogger(__name__)
		self._callbacks = []

		self._queues = dict()
		for key, queue_factory in queue_factories.items():
			self._queues[key] = queue_factory(self._analysis_finished)

	def register_finish_callback(self, callback):
		self._callbacks.append(callback)

	def unregister_finish_callback(self, callback):
		self._callbacks.remove(callback)

	def enqueue(self, entry, high_priority=False):
		if entry is None:
			return False

		if not entry.type in self._queues:
			return False

		self._queues[entry.type].enqueue(entry, high_priority=high_priority)
		return True

	def dequeue(self, entry):
		if entry is None:
			return False

		if not entry.type in self._queues:
			return False

		self._queues[entry.type].dequeue(entry.location, entry.path)

	def dequeue_folder(self, destination, path):
		for q in self._queues.values():
			q.dequeue_folder(destination, path)

	def pause(self):
		for q in self._queues.values():
			q.pause()

	def resume(self):
		for q in self._queues.values():
			q.resume()

	def _analysis_finished(self, entry, result):
		for callback in self._callbacks:
			try:
				callback(entry, result)
			except:
				self._logger.exception(u"Error while pushing analysis data to callback {}".format(callback),
				                       extra=dict(callback=fqcn(callback)))
		eventManager().fire(Events.METADATA_ANALYSIS_FINISHED, {"name": entry.name,
		                                                        "path": entry.path,
		                                                        "origin": entry.location,
		                                                        "result": result})

class AbstractAnalysisQueue(object):
	"""
	The :class:`AbstractAnalysisQueue` is the parent class of all specific analysis queues such as the
	:class:`GcodeAnalysisQueue`. It offers methods to enqueue new entries to analyze and pausing and resuming analysis
	processing.

	Arguments:
	    finished_callback (callable): Callback that will be called upon finishing analysis of an entry in the queue.
	        The callback will be called with the analyzed entry as the first argument and the analysis result as
	        returned from the queue implementation as the second parameter.

	.. automethod:: _do_analysis

	.. automethod:: _do_abort
	"""

	LOW_PRIO = 100
	LOW_PRIO_ABORTED = 75
	HIGH_PRIO = 50
	HIGH_PRIO_ABORTED = 0

	def __init__(self, finished_callback):
		self._logger = logging.getLogger(__name__)

		self._finished_callback = finished_callback

		self._active = threading.Event()
		self._active.set()

		self._done = threading.Event()
		self._done.clear()

		self._currentFile = None
		self._currentProgress = None

		self._queue = queue.PriorityQueue()
		self._current = None
		self._current_highprio = False

		self._worker = threading.Thread(target=self._work)
		self._worker.daemon = True
		self._worker.start()

	def enqueue(self, entry, high_priority=False):
		"""
		Enqueues an ``entry`` for analysis by the queue.

		If ``high_priority`` is True (defaults to False), the entry will be prioritized and hence processed before
		other entries in the queue with normal priority.

		Arguments:
		    entry (QueueEntry): The :class:`QueueEntry` to analyze.
		    high_priority (boolean): Whether to process the provided entry with high priority (True) or not
		        (False, default)
		"""

		if settings().get(["gcodeAnalysis", "runAt"]) == "never":
			self._logger.debug("Ignoring entry {entry} for analysis queue".format(entry=entry))
			return
		elif high_priority:
			self._logger.debug("Adding entry {entry} to analysis queue with high priority".format(entry=entry))
			prio = self.__class__.HIGH_PRIO
		else:
			self._logger.debug("Adding entry {entry} to analysis queue with low priority".format(entry=entry))
			prio = self.__class__.LOW_PRIO

		self._queue.put((prio, entry, high_priority))
		if high_priority and self._current is not None and not self._current_highprio:
			self._logger.debug("Aborting current analysis in favor of high priority one")
			self._do_abort()

	def dequeue(self, location, path):
		if self._current is not None and self._current.location == location \
				and self._current.path == path:
			self._do_abort(reenqueue=False)
			self._done.wait()
			self._done.clear()

	def dequeue_folder(self, location, path):
		if self._current is not None and self._current.location == location \
				and self._current.path.startswith(path + "/"):
			self._do_abort(reenqueue=False)
			self._done.wait()
			self._done.clear()

	def pause(self):
		"""
		Pauses processing of the queue, e.g. when a print is active.
		"""

		self._logger.debug("Pausing analysis")
		self._active.clear()
		if self._current is not None:
			self._logger.debug("Aborting running analysis, will restart when analyzer is resumed")
			self._do_abort()

	def resume(self):
		"""
		Resumes processing of the queue, e.g. when a print has finished.
		"""

		self._logger.debug("Resuming analyzer")
		self._active.set()

	def _work(self):
		while True:
			(priority, entry, high_priority) = self._queue.get()
			self._logger.debug("Processing entry {} from queue (priority {})".format(entry, priority))
			self._active.wait()

			try:
				self._analyze(entry, high_priority=high_priority)
				self._queue.task_done()
				self._done.set()
			except AnalysisAborted as ex:
				if ex.reenqueue:
					self._queue.put((self.__class__.HIGH_PRIO_ABORTED if high_priority else self.__class__.LOW_PRIO_ABORTED,
					                 entry,
					                 high_priority))
				self._logger.debug("Running analysis of entry {} aborted".format(entry))
				self._queue.task_done()
				self._done.set()
			else:
				time.sleep(1.0)

	def _analyze(self, entry, high_priority=False):
		path = entry.absolute_path
		if path is None or not os.path.exists(path):
			return

		self._current = entry
		self._current_highprio = high_priority
		self._current_progress = 0

		try:
			start_time = monotonic_time()
			self._logger.info("Starting analysis of {}".format(entry))
			eventManager().fire(Events.METADATA_ANALYSIS_STARTED, {"name": entry.name,
			                                                       "path": entry.path,
			                                                       "origin": entry.location,
			                                                       "type": entry.type})
			try:
				result = self._do_analysis(high_priority=high_priority)
			except TypeError:
				result = self._do_analysis()
			self._logger.info("Analysis of entry {} finished, needed {:.2f}s".format(entry, monotonic_time() - start_time))
			self._finished_callback(self._current, result)
		finally:
			self._current = None
			self._current_progress = None

	def _do_analysis(self, high_priority=False):
		"""
		Performs the actual analysis of the current entry which can be accessed via ``self._current``. Needs to be
		overridden by sub classes.

		Arguments:
		    high_priority (bool): Whether the current entry has high priority or not.

		Returns:
		    object: The result of the analysis which will be forwarded to the ``finished_callback`` provided during
		        construction.
		"""
		return None

	def _do_abort(self, reenqueue=True):
		"""
		Aborts analysis of the current entry. Needs to be overridden by sub classes.
		"""
		pass


class GcodeAnalysisQueue(AbstractAnalysisQueue):
	"""
	A queue to analyze GCODE files. Analysis results are :class:`dict` instances structured as follows:

	.. list-table::
	   :widths: 25 70

	   - * **Key**
	     * **Description**
	   - * ``estimatedPrintTime``
	     * Estimated time the file take to print, in seconds
	   - * ``filament``
	     * Substructure describing estimated filament usage. Keys are ``tool0`` for the first extruder, ``tool1`` for
	       the second and so on. For each tool extruded length and volume (based on diameter) are provided.
	   - * ``filament.toolX.length``
	     * The extruded length in mm
	   - * ``filament.toolX.volume``
	     * The extruded volume in cm³
	   - * ``printingArea``
	     * Bounding box of the printed object in the print volume (minimum and maximum coordinates)
	   - * ``printingArea.minX``
	     * Minimum X coordinate of the printed object
	   - * ``printingArea.maxX``
	     * Maximum X coordinate of the printed object
	   - * ``printingArea.minY``
	     * Minimum Y coordinate of the printed object
	   - * ``printingArea.maxY``
	     * Maximum Y coordinate of the printed object
	   - * ``printingArea.minZ``
	     * Minimum Z coordinate of the printed object
	   - * ``printingArea.maxZ``
	     * Maximum Z coordinate of the printed object
	   - * ``dimensions``
	     * Dimensions of the printed object in X, Y, Z
	   - * ``dimensions.width``
	     * Width of the printed model along the X axis, in mm
	   - * ``dimensions.depth``
	     * Depth of the printed model along the Y axis, in mm
	   - * ``dimensions.height``
	     * Height of the printed model along the Z axis, in mm
	"""

	def __init__(self, finished_callback):
		AbstractAnalysisQueue.__init__(self, finished_callback)

		self._aborted = False
		self._reenqueue = False

	def _do_analysis(self, high_priority=False):
		import sarge
		import sys
		import yaml

		if self._current.analysis:
			return self._current.analysis
		try:
			throttle = settings().getFloat(["gcodeAnalysis", "throttle_highprio"]) if high_priority \
				else settings().getFloat(["gcodeAnalysis", "throttle_normalprio"])
			throttle_lines = settings().getInt(["gcodeAnalysis", "throttle_lines"])
			max_extruders = settings().getInt(["gcodeAnalysis", "maxExtruders"])
			g90_extruder = settings().getBoolean(["feature", "g90InfluencesExtruder"])
			speedx = self._current.printer_profile["axes"]["x"]["speed"]
			speedy = self._current.printer_profile["axes"]["y"]["speed"]
			offsets = self._current.printer_profile["extruder"]["offsets"]

			command = [sys.executable, "-m", "octoprint", "analysis", "gcode",
			           "--speed-x={}".format(speedx), "--speed-y={}".format(speedy),
			           "--max-t={}".format(max_extruders), "--throttle={}".format(throttle),
			           "--throttle-lines={}".format(throttle_lines)]
			for offset in offsets[1:]:
				command += ["--offset", str(offset[0]), str(offset[1])]
			if g90_extruder:
				command += ["--g90-extruder"]
			command.append(self._current.absolute_path)

			self._logger.info("Invoking analysis command: {}".format(" ".join(command)))

			self._aborted = False
			p = sarge.run(command, async_=True, stdout=sarge.Capture())

			while len(p.commands) == 0:
				# somewhat ugly... we can't use wait_events because
				# the events might not be all set if an exception
				# by sarge is triggered within the async process
				# thread
				time.sleep(0.01)

			# by now we should have a command, let's wait for its
			# process to have been prepared
			p.commands[0].process_ready.wait()

			if not p.commands[0].process:
				# the process might have been set to None in case of any exception
				raise RuntimeError("Error while trying to run command {}".format(" ".join(command)))

			try:
				# let's wait for stuff to finish
				while p.returncode is None:
					if self._aborted:
						# oh, we shall abort, let's do so!
						p.commands[0].terminate()
						raise AnalysisAborted(reenqueue=self._reenqueue)

					# else continue
					p.commands[0].poll()
			finally:
				p.close()

			output = p.stdout.text
			self._logger.debug("Got output: {!r}".format(output))

			if not "RESULTS:" in output:
				raise RuntimeError("No analysis result found")

			_, output = output.split("RESULTS:")
			analysis = yaml.safe_load(output)

			result = dict()
			result["printingArea"] = analysis["printing_area"]
			result["dimensions"] = analysis["dimensions"]
			if analysis["total_time"]:
				result["estimatedPrintTime"] = analysis["total_time"] * 60
			if analysis["extrusion_length"]:
				result["filament"] = dict()
				for i in range(len(analysis["extrusion_length"])):
					result["filament"]["tool%d" % i] = {
						"length": analysis["extrusion_length"][i],
						"volume": analysis["extrusion_volume"][i]
					}
			return result
		finally:
			self._gcode = None

	def _do_abort(self, reenqueue=True):
		self._aborted = True
		self._reenqueue = reenqueue



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"


import time

from octoprint.util import monotonic_time

def can_perform_update(target, check, online=True):
	return True


def perform_update(target, check, target_version, log_cb=None, online=True):
	duration = check.get("duration", 30)

	now = monotonic_time()
	end = now + duration
	while now < end:
		log_cb(["{}s left...".format(end - now)], prefix=">", stream="output")
		time.sleep(5)
		now = monotonic_time()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


import sarge
import logging
import re
import time

from . import to_unicode, to_native_str
from .platform import get_os

from past.builtins import unicode

try:
	import queue
except ImportError:  # pragma: no cover
	import Queue as queue

# These regexes are based on the colorama package
# Author: Jonathan Hartley
# License: BSD-3 (https://github.com/tartley/colorama/blob/master/LICENSE.txt)
# Website: https://github.com/tartley/colorama/
_ANSI_CSI_PATTERN = b"\001?\033\[(\??(?:\d|;)*)([a-zA-Z])\002?"  # Control Sequence Introducer
_ANSI_OSC_PATTERN = b"\001?\033\]((?:.|;)*?)(\x07)\002?"         # Operating System Command
_ANSI_REGEX = re.compile(b"|".join([_ANSI_CSI_PATTERN,
                                    _ANSI_OSC_PATTERN]))


def clean_ansi(line):
	"""
	Removes ANSI control codes from ``line``.

	Parameters:
	    line (bytes or unicode): the line to process

	Returns:
	    (bytes or unicode) The line without any ANSI control codes

	Example::

	    >>> text = b"Some text with some \x1b[31mred words\x1b[39m in it"
	    >>> clean_ansi(text) # doctest: +ALLOW_BYTES
	    'Some text with some red words in it'
	    >>> text = b"We \x1b[?25lhide the cursor here and then \x1b[?25hshow it again here"
	    >>> clean_ansi(text) # doctest: +ALLOW_BYTES
	    'We hide the cursor here and then show it again here'
	"""
	if isinstance(line, unicode):
		return _ANSI_REGEX.sub(b"", line.encode("latin1")).decode("latin1")
	return _ANSI_REGEX.sub(b"", line)


class CommandlineError(Exception):
	"""
	Raised by :py:func:`~octoprint.util.commandline.CommandLineCaller.checked_call` on non zero return codes

	Arguments:
	    returncode (int): the return code of the command
	    stdout (str): the stdout output produced by the command
	    stderr (str): the stderr output produced by the command
	"""
	def __init__(self, returncode, stdout, stderr):
		self.returncode = returncode
		self.stdout = stdout
		self.stderr = stderr


class CommandlineCaller(object):
	"""
	The CommandlineCaller is a utility class that allows running command line commands while logging their stdout
	and stderr via configurable callback functions.

	Callbacks are expected to have a signature matching

	.. code-block:: python

	   def callback(*lines):
	       do_something_with_the_passed_lines()

	The class utilizes sarge underneath.

	Example:

	.. code-block:: python

	   from octoprint.util.commandline import CommandLineCaller, CommandLineError

	   def log(prefix, *lines):
	       for line in lines:
	           print("{} {}".format(prefix, line))

	   def log_stdout(*lines):
	       log(">>>", *lines)

	   def log_stderr(*lines):
	       log("!!!", *lines)

	   def log_call(*lines)
	       log("---", *lines)

	   caller = CommandLineCaller()
	   caller.on_log_call = log_call
	   caller.on_log_stdout = log_stdout
	   caller.on_log_stderr = log_stderr

	   try:
	       caller.checked_call(["some", "command", "with", "parameters"])
	   except CommandLineError as err:
	       print("Command returned {}".format(err.returncode))
	   else:
	       print("Command finished successfully")
	"""

	def __init__(self):
		self._logger = logging.getLogger(__name__)

		self.on_log_call = lambda *args, **kwargs: None
		"""Callback for the called command line"""

		self.on_log_stdout = lambda *args, **kwargs: None
		"""Callback for stdout output"""

		self.on_log_stderr = lambda *args, **kwargs: None
		"""Callback for stderr output"""

	def checked_call(self, command, **kwargs):
		"""
		Calls a command and raises an error if it doesn't return with return code 0

		Args:
		    command (list, tuple or str): command to call
		    kwargs (dict): additional keyword arguments to pass to the sarge ``run`` call (note that ``_async``,
		                   ``stdout`` and ``stderr`` will be overwritten)

		Returns:
		    (tuple) a 3-tuple of return code, full stdout and full stderr output

		Raises:
		    CommandlineError
		"""
		returncode, stdout, stderr = self.call(command, **kwargs)

		if returncode != 0:
			raise CommandlineError(returncode, stdout, stderr)

		return returncode, stdout, stderr

	def call(self, command, **kwargs):
		"""
		Calls a command

		Args:
		    command (list, tuple or str): command to call
		    kwargs (dict): additional keyword arguments to pass to the sarge ``run`` call (note that ``_async``,
		                   ``stdout`` and ``stderr`` will be overwritten)

		Returns:
		    (tuple) a 3-tuple of return code, full stdout and full stderr output
		"""

		if isinstance(command, (list, tuple)):
			joined_command = " ".join(command)
		else:
			joined_command = command
		self._logger.debug("Calling: {}".format(joined_command))
		self.on_log_call(joined_command)

		# if we are running under windows, make sure there are no unicode strings in the env
		if get_os() == "windows" and "env" in kwargs:
			kwargs["env"] = dict((k, to_native_str(v)) for k, v in kwargs["env"].items())

		delimiter = kwargs.get("delimiter", b'\n')
		try:
			kwargs.pop("delimiter")
		except KeyError:
			pass

		buffer_size = kwargs.get("buffer_size", -1)
		try:
			kwargs.pop("buffer_size")
		except KeyError:
			pass

		kwargs.update(dict(async_=True,
		                   stdout=DelimiterCapture(delimiter=delimiter, buffer_size=buffer_size),
		                   stderr=DelimiterCapture(delimiter=delimiter, buffer_size=buffer_size)))

		p = sarge.run(command, **kwargs)
		while len(p.commands) == 0:
			# somewhat ugly... we can't use wait_events because
			# the events might not be all set if an exception
			# by sarge is triggered within the async process
			# thread
			time.sleep(0.01)

		# by now we should have a command, let's wait for its
		# process to have been prepared
		p.commands[0].process_ready.wait()

		if not p.commands[0].process:
			# the process might have been set to None in case of any exception
			self._logger.error("Error while trying to run command {}".format(joined_command))
			return None, [], []

		all_stdout = []
		all_stderr = []

		def process_lines(lines, logger):
			if not lines:
				return []
			l = self._preprocess_lines(*map(lambda x: to_unicode(x, errors="replace"), lines))
			logger(*l)
			return list(l)

		def process_stdout(lines):
			return process_lines(lines, self._log_stdout)

		def process_stderr(lines):
			return process_lines(lines, self._log_stderr)

		try:
			while p.returncode is None:
				all_stderr += process_stderr(p.stderr.readlines(timeout=0.5))
				all_stdout += process_stdout(p.stdout.readlines(timeout=0.5))
				p.commands[0].poll()

		finally:
			p.close()

		all_stderr += process_stderr(p.stderr.readlines())
		all_stdout += process_stdout(p.stdout.readlines())

		return p.returncode, all_stdout, all_stderr

	def _log_stdout(self, *lines):
		self.on_log_stdout(*lines)

	def _log_stderr(self, *lines):
		self.on_log_stderr(*lines)

	def _preprocess_lines(self, *lines):
		return lines


class DelimiterCapture(sarge.Capture):
	def __init__(self, delimiter=b'\n', *args, **kwargs):
		self._delimiter = delimiter
		sarge.Capture.__init__(self, *args, **kwargs)

	def readline(self, size=-1, block=True, timeout=None):
		if not self.streams_open():
			block = False
			timeout = None
		else:
			timeout = timeout or self.timeout
		if self.current is None:
			try:
				self.current = self.buffer.get(block, timeout)
			except queue.Empty:
				self.current = b''
		while self._delimiter not in self.current:
			try:
				self.current += self.buffer.get(block, timeout)
			except queue.Empty:
				break
		if self._delimiter not in self.current:
			result = self.current
			self.current = None
		else:
			i = self.current.index(self._delimiter)
			if 0 < size < i:
				i = size - 1
			result = self.current[:i + 1]
			self.current = self.current[i + 1:]
		return result



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'

class FileDestinations(object):

	SDCARD = "sdcard"
	LOCAL = "local"



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import io
import os
import tarfile
import zipfile
import logging

try:
	from os import scandir
except ImportError:
	from scandir import scandir

from collections import defaultdict

from flask import request, jsonify, make_response

from octoprint.settings import settings

from octoprint.server.api import api
from octoprint.server.util.flask import no_firstrun_access

from octoprint.plugin import plugin_manager
from octoprint.access.permissions import Permissions

from flask_babel import Locale

@api.route("/languages", methods=["GET"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def getInstalledLanguagePacks():
	translation_folder = settings().getBaseFolder("translations", check_writable=False)
	if not os.path.exists(translation_folder):
		return jsonify(language_packs=dict(_core=[]))

	core_packs = []
	plugin_packs = defaultdict(lambda: dict(identifier=None, display=None, languages=[]))
	for entry in scandir(translation_folder):
		if not entry.is_dir():
			continue

		def load_meta(path, locale):
			meta = dict()

			meta_path = os.path.join(path, "meta.yaml")
			if os.path.isfile(meta_path):
				import yaml
				try:
					with io.open(meta_path, 'rt', encoding='utf-8') as f:
						meta = yaml.safe_load(f)
				except Exception:
					logging.getLogger(__name__).exception("Could not load %s", meta_path)
					pass
				else:
					import datetime
					if "last_update" in meta and isinstance(meta["last_update"], datetime.datetime):
						meta["last_update"] = (meta["last_update"] - datetime.datetime(1970,1,1)).total_seconds()

			l = Locale.parse(locale)
			meta["locale"] = locale
			meta["locale_display"] = l.display_name
			meta["locale_english"] = l.english_name
			return meta

		if entry.name == "_plugins":
			for plugin_entry in scandir(entry.path):
				if not plugin_entry.is_dir():
					continue

				if not plugin_entry.name in plugin_manager().plugins:
					continue

				plugin_info = plugin_manager().plugins[plugin_entry.name]

				plugin_packs[plugin_entry.name]["identifier"] = plugin_entry.name
				plugin_packs[plugin_entry.name]["display"] = plugin_info.name

				for language_entry in scandir(plugin_entry.path):
					try:
						plugin_packs[plugin_entry.name]["languages"].append(load_meta(language_entry.path, language_entry.name))
					except Exception:
						logging.getLogger(__name__).exception("Error while parsing metadata for language pack {} from {} for plugin {}".format(language_entry.name,
						                                                                                                                       language_entry.path,
						                                                                                                                       plugin_entry.name))
						continue
		else:
			try:
				core_packs.append(load_meta(entry.path, entry.name))
			except ValueError:
				logging.getLogger(__name__).exception("Core language pack {} doesn't appear to actually be one".format(entry.name))
			except Exception:
				logging.getLogger(__name__).exception("Error while parsing metadata for core language pack {} from {}".format(entry.name,
				                                                                                                              entry.path))

	result = dict(_core=dict(identifier="_core", display="Core", languages=core_packs))
	result.update(plugin_packs)
	return jsonify(language_packs=result)

@api.route("/languages", methods=["POST"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def uploadLanguagePack():
	input_name = "file"
	input_upload_path = input_name + "." + settings().get(["server", "uploads", "pathSuffix"])
	input_upload_name = input_name + "." + settings().get(["server", "uploads", "nameSuffix"])
	if not input_upload_path in request.values or not input_upload_name in request.values:
		return make_response("No file included", 400)

	upload_name = request.values[input_upload_name]
	upload_path = request.values[input_upload_path]

	exts = filter(lambda x: upload_name.lower().endswith(x), (".zip", ".tar.gz", ".tgz", ".tar"))
	if not len(exts):
		return make_response("File doesn't have a valid extension for a language pack archive", 400)

	target_path = settings().getBaseFolder("translations")

	if tarfile.is_tarfile(upload_path):
		_unpack_uploaded_tarball(upload_path, target_path)
	elif zipfile.is_zipfile(upload_path):
		_unpack_uploaded_zipfile(upload_path, target_path)
	else:
		return make_response("Neither zip file nor tarball included", 400)

	return getInstalledLanguagePacks()

@api.route("/languages/<string:locale>/<string:pack>", methods=["DELETE"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def deleteInstalledLanguagePack(locale, pack):

	if pack == "_core":
		target_path = os.path.join(settings().getBaseFolder("translations"), locale)
	else:
		target_path = os.path.join(settings().getBaseFolder("translations"), "_plugins", pack, locale)

	if os.path.isdir(target_path):
		import shutil
		shutil.rmtree(target_path)

	return getInstalledLanguagePacks()

def _unpack_uploaded_zipfile(path, target):
	with zipfile.ZipFile(path, "r") as zip:
		# sanity check
		map(_validate_archive_name, zip.namelist())

		# unpack everything
		zip.extractall(target)

def _unpack_uploaded_tarball(path, target):
	with tarfile.open(path, "r") as tar:
		# sanity check
		map(_validate_archive_name, tar.getmembers())

		# unpack everything
		tar.extractall(target)

def _validate_archive_name(name):
	if name.startswith("/") or ".." in name:
		raise InvalidLanguagePack("Provided language pack contains invalid name {name}".format(**locals()))


class InvalidLanguagePack(Exception):
	pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

import sys

try:
	import fcntl
except ImportError:
	fcntl = None

# set_close_exec

if fcntl is not None and hasattr(fcntl, "FD_CLOEXEC"):
	def set_close_exec(handle):
		flags = fcntl.fcntl(handle, fcntl.F_GETFD)
		flags |= fcntl.FD_CLOEXEC
		fcntl.fcntl(handle, fcntl.F_SETFD, flags)

elif sys.platform == "win32":
	def set_close_exec(handle):
		import ctypes
		import ctypes.wintypes

		# see https://msdn.microsoft.com/en-us/library/ms724935(v=vs.85).aspx
		SetHandleInformation = ctypes.windll.kernel32.SetHandleInformation
		SetHandleInformation.argtypes = (ctypes.wintypes.HANDLE, ctypes.wintypes.DWORD, ctypes.wintypes.DWORD)
		SetHandleInformation.restype = ctypes.c_bool

		HANDLE_FLAG_INHERIT = 0x00000001

		result = SetHandleInformation(handle, HANDLE_FLAG_INHERIT, 0)
		if not result:
			raise ctypes.GetLastError()

else:
	def set_close_exec(handle):
		# no-op
		pass


# current os

_OPERATING_SYSTEMS = dict(windows=["win32"],
                          linux=lambda x: x.startswith("linux"),
                          macos=["darwin"],
                          freebsd=lambda x: x.startswith("freebsd"))
OPERATING_SYSTEM_UNMAPPED = "unmapped"


def get_os():
	for identifier, platforms in _OPERATING_SYSTEMS.items():
		if (callable(platforms) and platforms(sys.platform)) or (isinstance(platforms, list) and sys.platform in platforms):
			return identifier
	else:
		return OPERATING_SYSTEM_UNMAPPED


def is_os_compatible(compatibility_entries, current_os=None):
	"""
	Tests if the ``current_os`` or ``sys.platform`` are blacklisted or whitelisted in ``compatibility_entries``
	"""
	if len(compatibility_entries) == 0:
		# shortcut - no compatibility info means we are compatible
		return True

	if current_os is None:
		current_os = get_os()

	negative_entries = list(map(lambda x: x[1:], filter(lambda x: x.startswith("!"), compatibility_entries)))
	positive_entries = list(filter(lambda x: not x.startswith("!"), compatibility_entries))

	negative_match = False
	if negative_entries:
		# check if we are blacklisted
		negative_match = current_os in negative_entries or any(map(lambda x: sys.platform.startswith(x), negative_entries))

	positive_match = True
	if positive_entries:
		# check if we are whitelisted
		positive_match = current_os in positive_entries or any(map(lambda x: sys.platform.startswith(x), positive_entries))

	return positive_match and not negative_match




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import io
import logging
import os
import shutil
import re
import pylru
import copy

try:
	from os import scandir, walk
except ImportError:
	from scandir import scandir, walk

from octoprint.util import atomic_write
from contextlib import contextmanager

from past.builtins import basestring

from emoji import demojize
from slugify import Slugify

import octoprint.filemanager

from octoprint.util import is_hidden_path, to_unicode, to_bytes

class StorageInterface(object):
	"""
	Interface of storage adapters for OctoPrint.
	"""

	@property
	def analysis_backlog(self):
		"""
		Get an iterator over all items stored in the storage that need to be analysed by the :class:`~octoprint.filemanager.AnalysisQueue`.

		The yielded elements are expected as storage specific absolute paths to the respective files. Don't forget
		to recurse into folders if your storage adapter supports those.

		:return: an iterator yielding all un-analysed files in the storage
		"""
		# empty generator pattern, yield is intentionally unreachable
		return
		yield

	def analysis_backlog_for_path(self, path=None):
		# empty generator pattern, yield is intentionally unreachable
		return
		yield

	def last_modified(self, path=None, recursive=False):
		"""
		Get the last modification date of the specified ``path`` or ``path``'s subtree.

		Args:
		    path (str or None): Path for which to determine the subtree's last modification date. If left out or
		        set to None, defatuls to storage root.
		    recursive (bool): Whether to determine only the date of the specified ``path`` (False, default) or
		        the whole ``path``'s subtree (True).

		Returns: (float) The last modification date of the indicated subtree
		"""
		raise NotImplementedError()

	def file_in_path(self, path, filepath):
		"""
		Returns whether the file indicated by ``file`` is inside ``path`` or not.
		:param string path: the path to check
		:param string filepath: path to the file
		:return: ``True`` if the file is inside the path, ``False`` otherwise
		"""
		return NotImplementedError()

	def file_exists(self, path):
		"""
		Returns whether the file indicated by ``path`` exists or not.
		:param string path: the path to check for existence
		:return: ``True`` if the file exists, ``False`` otherwise
		"""
		raise NotImplementedError()

	def folder_exists(self, path):
		"""
		Returns whether the folder indicated by ``path`` exists or not.
		:param string path: the path to check for existence
		:return: ``True`` if the folder exists, ``False`` otherwise
		"""
		raise NotImplementedError()

	def list_files(self, path=None, filter=None, recursive=True):
		"""
		List all files in storage starting at ``path``. If ``recursive`` is set to True (the default), also dives into
		subfolders.

		An optional filter function can be supplied which will be called with a file name and file data and which has
		to return True if the file is to be included in the result or False if not.

		The data structure of the returned result will be a dictionary mapping from file names to entry data. File nodes
		will contain their metadata here, folder nodes will contain their contained files and folders. Example::

		   {
		     "some_folder": {
		       "name": "some_folder",
		       "path": "some_folder",
		       "type": "folder",
		       "children": {
		         "some_sub_folder": {
		           "name": "some_sub_folder",
		           "path": "some_folder/some_sub_folder",
		           "type": "folder",
		           "typePath": ["folder"],
		           "children": { ... }
		         },
		         "some_file.gcode": {
		           "name": "some_file.gcode",
		           "path": "some_folder/some_file.gcode",
		           "type": "machinecode",
		           "typePath": ["machinecode", "gcode"],
		           "hash": "<sha1 hash>",
		           "links": [ ... ],
		           ...
		         },
		         ...
		       }
		     "test.gcode": {
		       "name": "test.gcode",
		       "path": "test.gcode",
		       "type": "machinecode",
		       "typePath": ["machinecode", "gcode"],
		       "hash": "<sha1 hash>",
		       "links": [...],
		       ...
		     },
		     "test.stl": {
		       "name": "test.stl",
		       "path": "test.stl",
		       "type": "model",
		       "typePath": ["model", "stl"],
		       "hash": "<sha1 hash>",
		       "links": [...],
		       ...
		     },
		     ...
		   }

		:param string path:     base path from which to recursively list all files, optional, if not supplied listing will start
		                        from root of base folder
		:param function filter: a filter that matches the files that are to be returned, may be left out in which case no
		                        filtering will take place
		:param bool recursive:  will also step into sub folders for building the complete list if set to True, otherwise will only
		                        do one step down into sub folders to be able to populate the ``children``.
		:return: a dictionary mapping entry names to entry data that represents the whole file list
		"""
		raise NotImplementedError()

	def add_folder(self, path, ignore_existing=True, display=None):
		"""
		Adds a folder as ``path``

		The ``path`` will be sanitized.

		:param string path:          the path of the new folder
		:param bool ignore_existing: if set to True, no error will be raised if the folder to be added already exists
		:param unicode display:      display name of the folder
		:return: the sanitized name of the new folder to be used for future references to the folder
		"""
		raise NotImplementedError()

	def remove_folder(self, path, recursive=True):
		"""
		Removes the folder at ``path``

		:param string path:    the path of the folder to remove
		:param bool recursive: if set to True, contained folders and files will also be removed, otherwise an error will
		                       be raised if the folder is not empty (apart from any metadata files) when it's to be removed
		"""
		raise NotImplementedError()

	def copy_folder(self, source, destination):
		"""
		Copies the folder ``source`` to ``destination``

		:param string source: path to the source folder
		:param string destination: path to destination

		:return: the path in the storage to the copy of the folder
		"""
		raise NotImplementedError()

	def move_folder(self, source, destination):
		"""
		Moves the folder ``source`` to ``destination``

		:param string source: path to the source folder
		:param string destination: path to destination

		:return: the new path in the storage to the folder
		"""
		raise NotImplementedError()

	def add_file(self, path, file_object, printer_profile=None, links=None, allow_overwrite=False, display=None):
		"""
		Adds the file ``file_object`` as ``path``

		:param string path:            the file's new path, will be sanitized
		:param object file_object:     a file object that provides a ``save`` method which will be called with the destination path
		                               where the object should then store its contents
		:param object printer_profile: the printer profile associated with this file (if any)
		:param list links:             any links to add with the file
		:param bool allow_overwrite:   if set to True no error will be raised if the file already exists and the existing file
		                               and its metadata will just be silently overwritten
		:param unicode display:        display name of the file
		:return: the sanitized name of the file to be used for future references to it
		"""
		raise NotImplementedError()

	def remove_file(self, path):
		"""
		Removes the file at ``path``

		Will also take care of deleting the corresponding entries
		in the metadata and deleting all links pointing to the file.

		:param string path: path of the file to remove
		"""
		raise NotImplementedError()

	def copy_file(self, source, destination):
		"""
		Copies the file ``source`` to ``destination``

		:param string source: path to the source file
		:param string destination: path to destination

		:return: the path in the storage to the copy of the file
		"""
		raise NotImplementedError()

	def move_file(self, source, destination):
		"""
		Moves the file ``source`` to ``destination``

		:param string source: path to the source file
		:param string destination: path to destination

		:return: the new path in the storage to the file
		"""
		raise NotImplementedError()

	def has_analysis(self, path):
		"""
		Returns whether the file at path has been analysed yet

		:param path: virtual path to the file for which to retrieve the metadata
		"""
		raise NotImplementedError()

	def get_metadata(self, path):
		"""
		Retrieves the metadata for the file ``path``.

		:param path: virtual path to the file for which to retrieve the metadata
		:return: the metadata associated with the file
		"""
		raise NotImplementedError()

	def add_link(self, path, rel, data):
		"""
		Adds a link of relation ``rel`` to file ``path`` with the given ``data``.

		The following relation types are currently supported:

		  * ``model``: adds a link to a model from which the file was created/sliced, expected additional data is the ``name``
		    and optionally the ``hash`` of the file to link to. If the link can be resolved against another file on the
		    current ``path``, not only will it be added to the links of ``name`` but a reverse link of type ``machinecode``
		    referring to ``name`` and its hash will also be added to the linked ``model`` file
		  * ``machinecode``: adds a link to a file containing machine code created from the current file (model), expected
		    additional data is the ``name`` and optionally the ``hash`` of the file to link to. If the link can be resolved
		    against another file on the current ``path``, not only will it be added to the links of ``name`` but a reverse
		    link of type ``model`` referring to ``name`` and its hash will also be added to the linked ``model`` file.
		  * ``web``: adds a location on the web associated with this file (e.g. a website where to download a model),
		    expected additional data is a ``href`` attribute holding the website's URL and optionally a ``retrieved``
		    attribute describing when the content was retrieved

		Note that adding ``model`` links to files identifying as models or ``machinecode`` links to files identifying
		as machine code will be refused.

		:param path: path of the file for which to add a link
		:param rel: type of relation of the link to add (currently ``model``, ``machinecode`` and ``web`` are supported)
		:param data: additional data of the link to add
		"""
		raise NotImplementedError()

	def remove_link(self, path, rel, data):
		"""
		Removes the link consisting of ``rel`` and ``data`` from file ``name`` on ``path``.

		:param path: path of the file from which to remove the link
		:param rel: type of relation of the link to remove (currently ``model``, ``machinecode`` and ``web`` are supported)
		:param data: additional data of the link to remove, must match existing link
		"""
		raise NotImplementedError()

	def set_additional_metadata(self, path, key, data, overwrite=False, merge=False):
		"""
		Adds additional metadata to the metadata of ``path``. Metadata in ``data`` will be saved under ``key``.

		If ``overwrite`` is set and ``key`` already exists in ``name``'s metadata, the current value will be overwritten.

		If ``merge`` is set and ``key`` already exists and both ``data`` and the existing data under ``key`` are dictionaries,
		the two dictionaries will be merged recursively.

		:param path: the virtual path to the file for which to add additional metadata
		:param key: key of metadata to add
		:param data: metadata to add
		:param overwrite: if True and ``key`` already exists, it will be overwritten
		:param merge: if True and ``key`` already exists and both ``data`` and the existing data are dictionaries, they
		              will be merged
		"""
		raise NotImplementedError()

	def remove_additional_metadata(self, path, key):
		"""
		Removes additional metadata under ``key`` for ``name`` on ``path``

		:param path: the virtual path to the file for which to remove the metadata under ``key``
		:param key: the key to remove
		"""
		raise NotImplementedError()

	def canonicalize(self, path):
		"""
		Canonicalizes the given ``path``. The ``path`` may consist of both folder and file name, the underlying
		implementation must separate those if necessary.

		By default, this calls :func:`~octoprint.filemanager.StorageInterface.sanitize`, which also takes care
		of stripping any invalid characters.

		Args:
			path: the path to canonicalize

		Returns:
			a 2-tuple containing the canonicalized path and file name

		"""
		return self.sanitize(path)

	def sanitize(self, path):
		"""
		Sanitizes the given ``path``, stripping it of all invalid characters. The ``path`` may consist of both
		folder and file name, the underlying implementation must separate those if necessary and sanitize individually.

		:param string path: the path to sanitize
		:return: a 2-tuple containing the sanitized path and file name
		"""
		raise NotImplementedError()

	def sanitize_path(self, path):
		"""
		Sanitizes the given folder-only ``path``, stripping it of all invalid characters.
		:param string path: the path to sanitize
		:return: the sanitized path
		"""
		raise NotImplementedError()

	def sanitize_name(self, name):
		"""
		Sanitizes the given file ``name``, stripping it of all invalid characters.
		:param string name: the file name to sanitize
		:return: the sanitized name
		"""
		raise NotImplementedError()

	def split_path(self, path):
		"""
		Split ``path`` into base directory and file name.
		:param path: the path to split
		:return: a tuple (base directory, file name)
		"""
		raise NotImplementedError()

	def join_path(self, *path):
		"""
		Join path elements together
		:param path: path elements to join
		:return: joined representation of the path to be usable as fully qualified path for further operations
		"""
		raise NotImplementedError()

	def path_on_disk(self, path):
		"""
		Retrieves the path on disk for ``path``.

		Note: if the storage is not on disk and there exists no path on disk to refer to it, this method should
		raise an :class:`io.UnsupportedOperation`

		Opposite of :func:`path_in_storage`.

		:param string path: the virtual path for which to retrieve the path on disk
		:return: the path on disk to ``path``
		"""
		raise NotImplementedError()

	def path_in_storage(self, path):
		"""
		Retrieves the equivalent in the storage adapter for ``path``.

		Opposite of :func:`path_on_disk`.

		:param string path: the path for which to retrieve the storage path
		:return: the path in storage to ``path``
		"""
		raise NotImplementedError()


class StorageError(Exception):
	UNKNOWN = "unknown"
	INVALID_DIRECTORY = "invalid_directory"
	INVALID_FILE = "invalid_file"
	INVALID_SOURCE = "invalid_source"
	INVALID_DESTINATION = "invalid_destination"
	DOES_NOT_EXIST = "does_not_exist"
	ALREADY_EXISTS = "already_exists"
	SOURCE_EQUALS_DESTINATION = "source_equals_destination"
	NOT_EMPTY = "not_empty"

	def __init__(self, message, code=None, cause=None):
		BaseException.__init__(self)
		self.message = message
		self.cause = cause

		if code is None:
			code = StorageError.UNKNOWN
		self.code = code


class LocalFileStorage(StorageInterface):
	"""
	The ``LocalFileStorage`` is a storage implementation which holds all files, folders and metadata on disk.

	Metadata is managed inside ``.metadata.json`` files in the respective folders, indexed by the sanitized filenames
	stored within the folder. Metadata access is managed through an LRU cache to minimize access overhead.

	This storage type implements :func:`path_on_disk`.
	"""

	_UNICODE_VARIATIONS = re.compile("[\uFE00-\uFE0F]", re.U)

	@classmethod
	def _no_unicode_variations(cls, text):
		return cls._UNICODE_VARIATIONS.sub("", text)

	_SLUGIFY = Slugify()
	_SLUGIFY.safe_chars = "-_.()[] "

	@classmethod
	def _slugify(cls, text):
		text = to_unicode(text)
		text = cls._no_unicode_variations(text)
		text = demojize(text, delimiters=("", ""))
		return cls._SLUGIFY(text)

	def __init__(self, basefolder, create=False):
		"""
		Initializes a ``LocalFileStorage`` instance under the given ``basefolder``, creating the necessary folder
		if necessary and ``create`` is set to ``True``.

		:param string basefolder: the path to the folder under which to create the storage
		:param bool create:       ``True`` if the folder should be created if it doesn't exist yet, ``False`` otherwise
		"""
		self._logger = logging.getLogger(__name__)

		self.basefolder = os.path.realpath(os.path.abspath(to_unicode(basefolder)))
		if not os.path.exists(self.basefolder) and create:
			os.makedirs(self.basefolder)
		if not os.path.exists(self.basefolder) or not os.path.isdir(self.basefolder):
			raise StorageError("{basefolder} is not a valid directory".format(**locals()), code=StorageError.INVALID_DIRECTORY)

		import threading
		self._metadata_lock_mutex = threading.RLock()
		self._metadata_locks = dict()
		self._persisted_metadata_lock_mutex = threading.RLock()
		self._persisted_metadata_locks = dict()

		self._metadata_cache = pylru.lrucache(10)

		self._old_metadata = None
		self._initialize_metadata()

	def _initialize_metadata(self):
		self._logger.info("Initializing the file metadata for {}...".format(self.basefolder))

		old_metadata_path = os.path.join(self.basefolder, "metadata.yaml")
		backup_path = os.path.join(self.basefolder, "metadata.yaml.backup")

		if os.path.exists(old_metadata_path):
			# load the old metadata file
			try:
				with io.open(old_metadata_path, 'rt', encoding='utf-8') as f:
					import yaml
					self._old_metadata = yaml.safe_load(f)
			except Exception:
				self._logger.exception("Error while loading old metadata file")

			# make sure the metadata is initialized as far as possible
			self._list_folder(self.basefolder)

			# rename the old metadata file
			self._old_metadata = None
			try:
				import shutil
				shutil.move(old_metadata_path, backup_path)
			except Exception:
				self._logger.exception("Could not rename old metadata.yaml file")

		else:
			# make sure the metadata is initialized as far as possible
			self._list_folder(self.basefolder)

		self._logger.info("... file metadata for {} initialized successfully.".format(self.basefolder))

	@property
	def analysis_backlog(self):
		return self.analysis_backlog_for_path()

	def analysis_backlog_for_path(self, path=None):
		if path:
			path = self.sanitize_path(path)

		for entry in self._analysis_backlog_generator(path):
			yield entry

	def _analysis_backlog_generator(self, path=None):
		if path is None:
			path = self.basefolder

		metadata = self._get_metadata(path)
		if not metadata:
			metadata = dict()
		for entry in scandir(path):
			if is_hidden_path(entry.name):
				continue

			if entry.is_file() and octoprint.filemanager.valid_file_type(entry.name):
				if not entry.name in metadata or not isinstance(metadata[entry.name], dict) or not "analysis" in metadata[entry.name]:
					printer_profile_rels = self.get_link(entry.path, "printerprofile")
					if printer_profile_rels:
						printer_profile_id = printer_profile_rels[0]["id"]
					else:
						printer_profile_id = None

					yield entry.name, entry.path, printer_profile_id
			elif os.path.isdir(entry.path):
				for sub_entry in self._analysis_backlog_generator(entry.path):
					yield self.join_path(entry.name, sub_entry[0]), sub_entry[1], sub_entry[2]

	def last_modified(self, path=None, recursive=False):
		if path is None:
			path = self.basefolder
		else:
			path = os.path.join(self.basefolder, path)

		def last_modified_for_path(p):
			metadata = os.path.join(p, ".metadata.json")
			if os.path.exists(metadata):
				return max(os.stat(p).st_mtime, os.stat(metadata).st_mtime)
			else:
				return os.stat(p).st_mtime

		if recursive:
			return max(last_modified_for_path(root) for root, _, _ in walk(path))
		else:
			return last_modified_for_path(path)

	def file_in_path(self, path, filepath):
		filepath = self.sanitize_path(filepath)
		path = self.sanitize_path(path)

		return filepath == path or filepath.startswith(path + os.sep)

	def file_exists(self, path):
		path, name = self.sanitize(path)
		file_path = os.path.join(path, name)
		return os.path.exists(file_path) and os.path.isfile(file_path)

	def folder_exists(self, path):
		path, name = self.sanitize(path)
		folder_path = os.path.join(path, name)
		return os.path.exists(folder_path) and os.path.isdir(folder_path)

	def list_files(self, path=None, filter=None, recursive=True):
		if path:
			path = self.sanitize_path(to_unicode(path))
			base = self.path_in_storage(path)
			if base:
				base += "/"
		else:
			path = self.basefolder
			base = ""
		return self._list_folder(path, base=base, entry_filter=filter, recursive=recursive)

	def add_folder(self, path, ignore_existing=True, display=None):
		display_path, display_name = self.canonicalize(path)
		path = self.sanitize_path(display_path)
		name = self.sanitize_name(display_name)

		if display is not None:
			display_name = display

		folder_path = os.path.join(path, name)
		if os.path.exists(folder_path):
			if not ignore_existing:
				raise StorageError("{name} does already exist in {path}".format(**locals()), code=StorageError.ALREADY_EXISTS)
		else:
			os.mkdir(folder_path)

		if display_name != name:
			metadata = self._get_metadata_entry(path, name, default=dict())
			metadata["display"] = display_name
			self._update_metadata_entry(path, name, metadata)

		return self.path_in_storage((path, name))

	def remove_folder(self, path, recursive=True):
		path, name = self.sanitize(path)

		folder_path = os.path.join(path, name)
		if not os.path.exists(folder_path):
			return

		empty = True
		for entry in scandir(folder_path):
			if entry.name == ".metadata.json" or entry.name == ".metadata.yaml":
				continue
			empty = False
			break

		if not empty and not recursive:
			raise StorageError("{name} in {path} is not empty".format(**locals()), code=StorageError.NOT_EMPTY)

		import shutil
		shutil.rmtree(folder_path)

		self._remove_metadata_entry(path, name)

	def _get_source_destination_data(self, source, destination, must_not_equal=False):
		"""Prepares data dicts about source and destination for copy/move."""
		source_path, source_name = self.sanitize(source)

		destination_canon_path, destination_canon_name = self.canonicalize(destination)
		destination_path = self.sanitize_path(destination_canon_path)
		destination_name = self.sanitize_name(destination_canon_name)

		source_fullpath = os.path.join(source_path, source_name)
		destination_fullpath = os.path.join(destination_path, destination_name)

		if not os.path.exists(source_fullpath):
			raise StorageError("{} in {} does not exist".format(source_name, source_path), code=StorageError.INVALID_SOURCE)

		if not os.path.isdir(destination_path):
			raise StorageError("Destination path {} does not exist or is not a folder".format(destination_path), code=StorageError.INVALID_DESTINATION)
		if os.path.exists(destination_fullpath) and source_fullpath != destination_fullpath:
			raise StorageError("{} does already exist in {}".format(destination_name, destination_path), code=StorageError.INVALID_DESTINATION)

		source_meta = self._get_metadata_entry(source_path, source_name)
		if source_meta:
			source_display = source_meta.get("display", source_name)
		else:
			source_display = source_name

		if (must_not_equal or source_display == destination_canon_name) and source_fullpath == destination_fullpath:
			raise StorageError("Source {} and destination {} are the same folder".format(source_path, destination_path), code=StorageError.SOURCE_EQUALS_DESTINATION)

		source_data = dict(
			path=source_path,
			name=source_name,
			display=source_display,
			fullpath=source_fullpath,
		)
		destination_data = dict(
			path=destination_path,
			name=destination_name,
			display=destination_canon_name,
			fullpath=destination_fullpath,
		)
		return source_data, destination_data

	def _set_display_metadata(self, destination_data, source_data=None):
		if source_data and destination_data["name"] == source_data["name"] and source_data["name"] != source_data["display"]:
			display = source_data["display"]
		elif destination_data["name"] != destination_data["display"]:
			display = destination_data["display"]
		else:
			display = None

		destination_meta = self._get_metadata_entry(destination_data["path"], destination_data["name"],
		                                            default=dict())
		if display:
			destination_meta["display"] = display
			self._update_metadata_entry(destination_data["path"], destination_data["name"], destination_meta)
		elif "display" in destination_meta:
			del destination_meta["display"]
			self._update_metadata_entry(destination_data["path"], destination_data["name"], destination_meta)

	def copy_folder(self, source, destination):
		source_data, destination_data = self._get_source_destination_data(source, destination, must_not_equal=True)

		try:
			shutil.copytree(source_data["fullpath"], destination_data["fullpath"])
		except Exception as e:
			raise StorageError("Could not copy %s in %s to %s in %s" % (source_data["name"], source_data["path"], destination_data["name"], destination_data["path"]), cause=e)

		self._set_display_metadata(destination_data, source_data=source_data)

		return self.path_in_storage(destination_data["fullpath"])

	def move_folder(self, source, destination):
		source_data, destination_data = self._get_source_destination_data(source, destination)

		# only a display rename? Update that and bail early
		if source_data["fullpath"] == destination_data["fullpath"]:
			self._set_display_metadata(destination_data)
			return self.path_in_storage(destination_data["fullpath"])

		try:
			shutil.move(source_data["fullpath"], destination_data["fullpath"])
		except Exception as e:
			raise StorageError("Could not move %s in %s to %s in %s" % (source_data["name"], source_data["path"], destination_data["name"], destination_data["path"]), cause=e)

		self._set_display_metadata(destination_data, source_data=source_data)
		self._remove_metadata_entry(source_data["path"], source_data["name"])
		self._delete_metadata(source_data["fullpath"])

		return self.path_in_storage(destination_data["fullpath"])

	def add_file(self, path, file_object, printer_profile=None, links=None, allow_overwrite=False, display=None):
		display_path, display_name = self.canonicalize(path)
		path = self.sanitize_path(display_path)
		name = self.sanitize_name(display_name)

		if display:
			display_name = display

		if not octoprint.filemanager.valid_file_type(name):
			raise StorageError("{name} is an unrecognized file type".format(**locals()), code=StorageError.INVALID_FILE)

		file_path = os.path.join(path, name)
		if os.path.exists(file_path) and not os.path.isfile(file_path):
			raise StorageError("{name} does already exist in {path} and is not a file".format(**locals()), code=StorageError.ALREADY_EXISTS)
		if os.path.exists(file_path) and not allow_overwrite:
			raise StorageError("{name} does already exist in {path} and overwriting is prohibited".format(**locals()), code=StorageError.ALREADY_EXISTS)

		# make sure folders exist
		if not os.path.exists(path):
			# TODO persist display names of path segments!
			os.makedirs(path)

		# save the file
		file_object.save(file_path)

		# save the file's hash to the metadata of the folder
		file_hash = self._create_hash(file_path)
		metadata = self._get_metadata_entry(path, name, default=dict())
		metadata_dirty = False
		if not "hash" in metadata or metadata["hash"] != file_hash:
			# hash changed -> throw away old metadata
			metadata = dict(hash=file_hash)
			metadata_dirty = True

		if not "display" in metadata and display_name != name:
			# display name is not the same as file name -> store in metadata
			metadata["display"] = display_name
			metadata_dirty = True

		if metadata_dirty:
			self._update_metadata_entry(path, name, metadata)

		# process any links that were also provided for adding to the file
		if not links:
			links = []

		if printer_profile is not None:
			links.append(("printerprofile", dict(id=printer_profile["id"], name=printer_profile["name"])))

		self._add_links(name, path, links)

		# touch the file to set last access and modification time to now
		os.utime(file_path, None)

		return self.path_in_storage((path, name))

	def remove_file(self, path):
		path, name = self.sanitize(path)

		file_path = os.path.join(path, name)
		if not os.path.exists(file_path):
			return
		if not os.path.isfile(file_path):
			raise StorageError("{name} in {path} is not a file".format(**locals()), code=StorageError.INVALID_FILE)

		try:
			os.remove(file_path)
		except Exception as e:
			raise StorageError("Could not delete {name} in {path}".format(**locals()), cause=e)

		self._remove_metadata_entry(path, name)

	def copy_file(self, source, destination):
		source_data, destination_data = self._get_source_destination_data(source, destination, must_not_equal=True)

		try:
			shutil.copy2(source_data["fullpath"], destination_data["fullpath"])
		except Exception as e:
			raise StorageError("Could not copy %s in %s to %s in %s" % (source_data["name"], source_data["path"], destination_data["name"], destination_data["path"]), cause=e)

		self._copy_metadata_entry(source_data["path"], source_data["name"],
		                          destination_data["path"], destination_data["name"])
		self._set_display_metadata(destination_data, source_data=source_data)

		return self.path_in_storage(destination_data["fullpath"])

	def move_file(self, source, destination, allow_overwrite=False):
		source_data, destination_data = self._get_source_destination_data(source, destination)

		# only a display rename? Update that and bail early
		if source_data["fullpath"] == destination_data["fullpath"]:
			self._set_display_metadata(destination_data)
			return self.path_in_storage(destination_data["fullpath"])

		try:
			shutil.move(source_data["fullpath"], destination_data["fullpath"])
		except Exception as e:
			raise StorageError("Could not move %s in %s to %s in %s" % (source_data["name"], source_data["path"], destination_data["name"], destination_data["path"]), cause=e)

		self._copy_metadata_entry(source_data["path"], source_data["name"],
		                          destination_data["path"], destination_data["name"],
		                          delete_source=True)
		self._set_display_metadata(destination_data, source_data=source_data)

		return self.path_in_storage(destination_data["fullpath"])

	def has_analysis(self, path):
		metadata = self.get_metadata(path)
		return "analysis" in metadata

	def get_metadata(self, path):
		path, name = self.sanitize(path)
		return self._get_metadata_entry(path, name)

	def get_link(self, path, rel):
		path, name = self.sanitize(path)
		return self._get_links(name, path, rel)

	def add_link(self, path, rel, data):
		path, name = self.sanitize(path)
		self._add_links(name, path, [(rel, data)])

	def remove_link(self, path, rel, data):
		path, name = self.sanitize(path)
		self._remove_links(name, path, [(rel, data)])

	def add_history(self, path, data):
		path, name = self.sanitize(path)
		self._add_history(name, path, data)

	def update_history(self, path, index, data):
		path, name = self.sanitize(path)
		self._update_history(name, path, index, data)

	def remove_history(self, path, index):
		path, name = self.sanitize(path)
		self._delete_history(name, path, index)

	def set_additional_metadata(self, path, key, data, overwrite=False, merge=False):
		path, name = self.sanitize(path)
		metadata = self._get_metadata(path)
		metadata_dirty = False

		if not name in metadata:
			return

		metadata = self._copied_metadata(metadata, name)

		if not key in metadata[name] or overwrite:
			metadata[name][key] = data
			metadata_dirty = True
		elif key in metadata[name] and isinstance(metadata[name][key], dict) and isinstance(data, dict) and merge:
			current_data = metadata[name][key]

			import octoprint.util
			new_data = octoprint.util.dict_merge(current_data, data)
			metadata[name][key] = new_data
			metadata_dirty = True

		if metadata_dirty:
			self._save_metadata(path, metadata)

	def remove_additional_metadata(self, path, key):
		path, name = self.sanitize(path)
		metadata = self._get_metadata(path)

		if not name in metadata:
			return

		if not key in metadata[name]:
			return

		metadata = self._copied_metadata(metadata, name)
		del metadata[name][key]
		self._save_metadata(path, metadata)

	def split_path(self, path):
		path = to_unicode(path)
		split = path.split("/")
		if len(split) == 1:
			return "", split[0]
		else:
			return self.join_path(*split[:-1]), split[-1]

	def join_path(self, *path):
		return "/".join(map(to_unicode, path))

	def sanitize(self, path):
		"""
		Returns a ``(path, name)`` tuple derived from the provided ``path``.

		``path`` may be:
		  * a storage path
		  * an absolute file system path
		  * a tuple or list containing all individual path elements
		  * a string representation of the path
		  * with or without a file name

		Note that for a ``path`` without a trailing slash the last part will be considered a file name and
		hence be returned at second position. If you only need to convert a folder path, be sure to
		include a trailing slash for a string ``path`` or an empty last element for a list ``path``.
		"""

		path, name = self.canonicalize(path)
		name = self.sanitize_name(name)
		path = self.sanitize_path(path)
		return path, name

	def canonicalize(self, path):
		name = None
		if isinstance(path, basestring):
			path = to_unicode(path)
			if path.startswith(self.basefolder):
				path = path[len(self.basefolder):]
			path = path.replace(os.path.sep, "/")
			path = path.split("/")
		if isinstance(path, (list, tuple)):
			if len(path) == 1:
				name = to_unicode(path[0])
				path = ""
			else:
				name = to_unicode(path[-1])
				path = self.join_path(*map(to_unicode, path[:-1]))
		if not path:
			path = ""

		return path, name

	def sanitize_name(self, name):
		"""
		Raises a :class:`ValueError` for a ``name`` containing ``/`` or ``\\``. Otherwise
		slugifies the given ``name`` by converting it to ASCII, leaving ``-``, ``_``, ``.``,
		``(``, and ``)`` as is.
		"""
		name = to_unicode(name)

		if name is None:
			return None

		if "/" in name or "\\" in name:
			raise ValueError("name must not contain / or \\")

		result = self._slugify(name).replace(" ", "_")
		if result and result != "." and result != ".." and result[0] == ".":
			# hidden files under *nix
			result = result[1:]
		return result

	def sanitize_path(self, path):
		"""
		Ensures that the on disk representation of ``path`` is located under the configured basefolder. Resolves all
		relative path elements (e.g. ``..``) and sanitizes folder names using :func:`sanitize_name`. Final path is the
		absolute path including leading ``basefolder`` path.
		"""
		path = to_unicode(path)

		if len(path):
			if path[0] == "/":
				path = path[1:]
			elif path[0] == "." and path[1] == "/":
				path = path[2:]

		path_elements = path.split("/")
		joined_path = self.basefolder
		for path_element in path_elements:
			joined_path = os.path.join(joined_path, self.sanitize_name(path_element))
		path = os.path.realpath(joined_path)
		if not path.startswith(self.basefolder):
			raise ValueError("path not contained in base folder: {path}".format(**locals()))
		return path

	def _sanitize_entry(self, entry, path, entry_path):
		entry = to_unicode(entry)
		sanitized = self.sanitize_name(entry)
		if sanitized != entry:
			# entry is not sanitized yet, let's take care of that
			sanitized_path = os.path.join(path, sanitized)
			sanitized_name, sanitized_ext = os.path.splitext(sanitized)

			counter = 1
			while os.path.exists(sanitized_path):
				counter += 1
				sanitized = self.sanitize_name("{}_({}){}".format(sanitized_name, counter, sanitized_ext))
				sanitized_path = os.path.join(path, sanitized)

			try:
				shutil.move(entry_path, sanitized_path)

				self._logger.info("Sanitized \"{}\" to \"{}\"".format(entry_path, sanitized_path))
				return sanitized, sanitized_path
			except Exception:
				self._logger.exception("Error while trying to rename \"{}\" to \"{}\", ignoring file".format(entry_path, sanitized_path))
				raise

		return entry, entry_path

	def path_in_storage(self, path):
		if isinstance(path, (tuple, list)):
			path = self.join_path(*path)
		if isinstance(path, basestring):
			path = to_unicode(path)
			if path.startswith(self.basefolder):
				path = path[len(self.basefolder):]
			path = path.replace(os.path.sep, "/")
		if path.startswith("/"):
			path = path[1:]

		return path

	def path_on_disk(self, path):
		path, name = self.sanitize(path)
		return os.path.join(path, name)

	##~~ internals

	def _add_history(self, name, path, data):
		metadata = self._copied_metadata(self._get_metadata(path), name)

		if not "hash" in metadata[name]:
			metadata[name]["hash"] = self._create_hash(os.path.join(path, name))

		if not "history" in metadata[name]:
			metadata[name]["history"] = []

		metadata[name]["history"].append(data)
		self._calculate_stats_from_history(name, path, metadata=metadata, save=False)
		self._save_metadata(path, metadata)

	def _update_history(self, name, path, index, data):
		metadata = self._get_metadata(path)

		if not name in metadata or not "history" in metadata[name]:
			return

		metadata = self._copied_metadata(metadata, name)

		try:
			metadata[name]["history"][index].update(data)
			self._calculate_stats_from_history(name, path, metadata=metadata, save=False)
			self._save_metadata(path, metadata)
		except IndexError:
			pass

	def _delete_history(self, name, path, index):
		metadata = self._get_metadata(path)

		if not name in metadata or not "history" in metadata[name]:
			return

		metadata = self._copied_metadata(metadata, name)

		try:
			del metadata[name]["history"][index]
			self._calculate_stats_from_history(name, path, metadata=metadata, save=False)
			self._save_metadata(path, metadata)
		except IndexError:
			pass

	def _calculate_stats_from_history(self, name, path, metadata=None, save=True):
		if metadata is None:
			metadata = self._copied_metadata(self._get_metadata(path), name)

		if not "history" in metadata[name]:
			return

		# collect data from history
		former_print_times = dict()
		last_print = dict()


		for history_entry in metadata[name]["history"]:
			if not "printTime" in history_entry or not "success" in history_entry or not history_entry["success"] or not "printerProfile" in history_entry:
				continue

			printer_profile = history_entry["printerProfile"]
			if not printer_profile:
				continue

			print_time = history_entry["printTime"]
			try:
				print_time = float(print_time)
			except Exception:
				self._logger.warning("Invalid print time value found in print history for {} in {}/.metadata.json: {!r}".format(name, path, print_time))
				continue

			if not printer_profile in former_print_times:
				former_print_times[printer_profile] = []
			former_print_times[printer_profile].append(print_time)

			if not printer_profile in last_print or last_print[printer_profile] is None or ("timestamp" in history_entry and history_entry["timestamp"] > last_print[printer_profile]["timestamp"]):
				last_print[printer_profile] = history_entry

		# calculate stats
		statistics = dict(averagePrintTime=dict(), lastPrintTime=dict())

		for printer_profile in former_print_times:
			if not former_print_times[printer_profile]:
				continue
			statistics["averagePrintTime"][printer_profile] = sum(former_print_times[printer_profile]) / float(len(former_print_times[printer_profile]))

		for printer_profile in last_print:
			if not last_print[printer_profile]:
				continue
			statistics["lastPrintTime"][printer_profile] = last_print[printer_profile]["printTime"]

		metadata[name]["statistics"] = statistics

		if save:
			self._save_metadata(path, metadata)

	def _get_links(self, name, path, searched_rel):
		metadata = self._get_metadata(path)
		result = []

		if not name in metadata:
			return result

		if not "links" in metadata[name]:
			return result

		for data in metadata[name]["links"]:
			if not "rel" in data or not data["rel"] == searched_rel:
				continue
			result.append(data)
		return result

	def _add_links(self, name, path, links):
		file_type = octoprint.filemanager.get_file_type(name)
		if file_type:
			file_type = file_type[0]

		metadata = self._copied_metadata(self._get_metadata(path), name)
		metadata_dirty = False

		if not "hash" in metadata[name]:
			metadata[name]["hash"] = self._create_hash(os.path.join(path, name))

		if not "links" in metadata[name]:
			metadata[name]["links"] = []

		for rel, data in links:
			if (rel == "model" or rel == "machinecode") and "name" in data:
				if file_type == "model" and rel == "model":
					# adding a model link to a model doesn't make sense
					return
				elif file_type == "machinecode" and rel == "machinecode":
					# adding a machinecode link to a machinecode doesn't make sense
					return

				ref_path = os.path.join(path, data["name"])
				if not os.path.exists(ref_path):
					# file doesn't exist, we won't create the link
					continue

				# fetch hash of target file
				if data["name"] in metadata and "hash" in metadata[data["name"]]:
					hash = metadata[data["name"]]["hash"]
				else:
					hash = self._create_hash(ref_path)
					if not data["name"] in metadata:
						metadata[data["name"]] = dict(
							hash=hash,
							links=[]
						)
					else:
						metadata[data["name"]]["hash"] = hash

				if "hash" in data and not data["hash"] == hash:
					# file doesn't have the correct hash, we won't create the link
					continue

				if not "links" in metadata[data["name"]]:
					metadata[data["name"]]["links"] = []

				# add reverse link to link target file
				metadata[data["name"]]["links"].append(
					dict(rel="machinecode" if rel == "model" else "model", name=name, hash=metadata[name]["hash"])
				)
				metadata_dirty = True

				link_dict = dict(
					rel=rel,
					name=data["name"],
					hash=hash
				)

			elif rel == "web" and "href" in data:
				link_dict = dict(
					rel=rel,
					href=data["href"]
				)
				if "retrieved" in data:
					link_dict["retrieved"] = data["retrieved"]

			else:
				continue

			if link_dict:
				metadata[name]["links"].append(link_dict)
				metadata_dirty = True

		if metadata_dirty:
			self._save_metadata(path, metadata)

	def _remove_links(self, name, path, links):
		metadata = self._copied_metadata(self._get_metadata(path), name)
		metadata_dirty = False

		hash = metadata[name].get("hash", self._create_hash(os.path.join(path, name)))

		for rel, data in links:
			if (rel == "model" or rel == "machinecode") and "name" in data:
				if data["name"] in metadata and "links" in metadata[data["name"]]:
					ref_rel = "model" if rel == "machinecode" else "machinecode"
					for link in metadata[data["name"]]["links"]:
						if link["rel"] == ref_rel and "name" in link and link["name"] == name and "hash" in link and link["hash"] == hash:
							metadata[data["name"]] = copy.deepcopy(metadata[data["name"]])
							metadata[data["name"]]["links"].remove(link)
							metadata_dirty = True

			if "links" in metadata[name]:
				for link in metadata[name]["links"]:
					if not link["rel"] == rel:
						continue

					matches = True
					for k, v in data.items():
						if not k in link or not link[k] == v:
							matches = False
							break

					if not matches:
						continue

					metadata[name]["links"].remove(link)
					metadata_dirty = True

		if metadata_dirty:
			self._save_metadata(path, metadata)

	def _list_folder(self, path, base="", entry_filter=None, recursive=True, include_children=True, **kwargs):
		if entry_filter is None:
			entry_filter = kwargs.get("filter", None)

		metadata = self._get_metadata(path)
		if not metadata:
			metadata = dict()
		metadata_dirty = False

		result = dict()
		for entry in scandir(path):
			if is_hidden_path(entry.name):
				# no hidden files and folders
				continue

			try:
				entry_name = entry_display = entry.name
				entry_path = entry.path
				entry_is_file = entry.is_file()
				entry_is_dir = entry.is_dir()
				entry_stat = entry.stat()
			except Exception:
				# error while trying to fetch file metadata, that might be thanks to file already having
				# been moved or deleted - ignore it and continue
				continue

			try:
				new_entry_name, new_entry_path = self._sanitize_entry(entry_name, path, entry_path)
				if entry_name != new_entry_name or entry_path != new_entry_path:
					entry_display = to_unicode(entry_name)
					entry_name = new_entry_name
					entry_path = new_entry_path
					entry_stat = os.stat(entry_path)
			except Exception:
				# error while trying to rename the file, we'll continue here and ignore it
				continue

			path_in_location = entry_name if not base else base + entry_name

			try:
				# file handling
				if entry_is_file:
					type_path = octoprint.filemanager.get_file_type(entry_name)
					if not type_path:
						# only supported extensions
						continue
					else:
						file_type = type_path[0]

					if entry_name in metadata and isinstance(metadata[entry_name], dict):
						entry_metadata = metadata[entry_name]
						if not "display" in entry_metadata and entry_display != entry_name:
							if not metadata_dirty:
								metadata = self._copied_metadata(metadata, entry_name)
							metadata[entry_name]["display"] = entry_display
							entry_metadata["display"] = entry_display
							metadata_dirty = True
					else:
						if not metadata_dirty:
							metadata = self._copied_metadata(metadata, entry_name)
						entry_metadata = self._add_basic_metadata(path, entry_name,
						                                          display_name=entry_display,
						                                          save=False,
						                                          metadata=metadata)
						metadata_dirty = True

					# TODO extract model hash from source if possible to recreate link

					if not entry_filter or entry_filter(entry_name, entry_metadata):
						# only add files passing the optional filter
						extended_entry_data = dict()
						extended_entry_data.update(entry_metadata)
						extended_entry_data["name"] = entry_name
						extended_entry_data["display"] = entry_metadata.get("display", entry_name)
						extended_entry_data["path"] = path_in_location
						extended_entry_data["type"] = file_type
						extended_entry_data["typePath"] = type_path
						stat = entry_stat
						if stat:
							extended_entry_data["size"] = stat.st_size
							extended_entry_data["date"] = int(stat.st_mtime)

						result[entry_name] = extended_entry_data

				# folder recursion
				elif entry_is_dir:
					if entry_name in metadata and isinstance(metadata[entry_name], dict):
						entry_metadata = metadata[entry_name]
						if not "display" in entry_metadata and entry_display != entry_name:
							if not metadata_dirty:
								metadata = self._copied_metadata(metadata, entry_name)
							metadata[entry_name]["display"] = entry_display
							entry_metadata["display"] = entry_display
							metadata_dirty = True
					elif entry_name != entry_display:
						if not metadata_dirty:
							metadata = self._copied_metadata(metadata, entry_name)
						entry_metadata = self._add_basic_metadata(path, entry_name,
						                                          display_name=entry_display,
						                                          save=False,
						                                          metadata=metadata)
						metadata_dirty = True
					else:
						entry_metadata = dict()

					entry_data = dict(
						name=entry_name,
						display=entry_metadata.get("display", entry_name),
						path=path_in_location,
						type="folder",
						typePath=["folder"]
					)
					if recursive:
						sub_result = self._list_folder(entry_path, base=path_in_location + "/", entry_filter=entry_filter,
						                               recursive=recursive)
						entry_data["children"] = sub_result
					elif include_children:
						sub_result = self._list_folder(entry_path, base=path_in_location + "/", entry_filter=entry_filter,
						                               recursive=False, include_children=False)
						entry_data["children"] = sub_result

					if not entry_filter or entry_filter(entry_name, entry_data):
						# only add folders passing the optional filter

						def get_size():
							total_size = 0
							for element in entry_data["children"].values():
								if "size" in element:
									total_size += element["size"]

							return total_size

						extended_entry_data = dict()
						extended_entry_data.update(entry_data)
						if recursive:
							extended_entry_data["size"] = get_size()

						result[entry_name] = extended_entry_data
			except Exception:
				# So something went wrong somewhere while processing this file entry - log that and continue
				self._logger.exception("Error while processing entry {}".format(entry_path))
				continue

		# TODO recreate links if we have metadata less entries

		# save metadata
		if metadata_dirty:
			self._save_metadata(path, metadata)

		return result

	def _add_basic_metadata(self, path, entry, display_name=None, additional_metadata=None, save=True, metadata=None):
		if additional_metadata is None:
			additional_metadata = dict()

		if metadata is None:
			metadata = self._get_metadata(path)

		entry_path = os.path.join(path, entry)

		if os.path.isfile(entry_path):
			entry_data = dict(
				hash=self._create_hash(os.path.join(path, entry)),
				links=[],
				notes=[]
			)
			if path == self.basefolder and self._old_metadata is not None and entry in self._old_metadata and "gcodeAnalysis" in self._old_metadata[entry]:
				# if there is still old metadata available and that contains an analysis for this file, use it!
				entry_data["analysis"] = self._old_metadata[entry]["gcodeAnalysis"]

		elif os.path.isdir(entry_path):
			entry_data = dict()

		else:
			return

		if display_name is not None and not display_name == entry:
			entry_data["display"] = display_name

		entry_data.update(additional_metadata)

		metadata = copy.copy(metadata)
		metadata[entry] = entry_data

		if save:
			self._save_metadata(path, metadata)

		return entry_data

	def _create_hash(self, path):
		import hashlib

		blocksize = 65536
		hash = hashlib.sha1()
		with io.open(path, 'rb') as f:
			buffer = f.read(blocksize)
			while len(buffer) > 0:
				hash.update(buffer)
				buffer = f.read(blocksize)

		return hash.hexdigest()

	def _get_metadata_entry(self, path, name, default=None):
		with self._get_metadata_lock(path):
			metadata = self._get_metadata(path)
			return metadata.get(name, default)

	def _remove_metadata_entry(self, path, name):
		with self._get_metadata_lock(path):
			metadata = self._get_metadata(path)
			if not name in metadata:
				return

			metadata = copy.copy(metadata)

			if "hash" in metadata[name]:
				hash = metadata[name]["hash"]
				for m in metadata.values():
					if not "links" in m:
						continue
					links_hash = lambda link: "hash" in link and link["hash"] == hash and "rel" in link and (link["rel"] == "model" or link["rel"] == "machinecode")
					m["links"] = [link for link in m["links"] if not links_hash(link)]

			del metadata[name]
			self._save_metadata(path, metadata)

	def _update_metadata_entry(self, path, name, data):
		with self._get_metadata_lock(path):
			metadata = copy.copy(self._get_metadata(path))
			metadata[name] = data
			self._save_metadata(path, metadata)

	def _copy_metadata_entry(self, source_path, source_name, destination_path, destination_name, delete_source=False, updates=None):
		with self._get_metadata_lock(source_path):
			source_data = self._get_metadata_entry(source_path, source_name, default=dict())
			if not source_data:
				return

			if delete_source:
				self._remove_metadata_entry(source_path, source_name)

		if updates is not None:
			source_data.update(updates)

		with self._get_metadata_lock(destination_path):
			self._update_metadata_entry(destination_path, destination_name, source_data)

	def _get_metadata(self, path, force=False):
		import json

		if not force:
			metadata = self._metadata_cache.get(path)
			if metadata:
				return metadata

		self._migrate_metadata(path)

		metadata_path = os.path.join(path, ".metadata.json")

		metadata = None
		with self._get_persisted_metadata_lock(path):
			if os.path.exists(metadata_path):
				with io.open(metadata_path, 'rt', encoding='utf-8') as f:
					try:
						metadata = json.load(f)
					except Exception:
						self._logger.exception("Error while reading .metadata.json from {path}".format(**locals()))

		def valid_json(value):
			try:
				json.dumps(value, allow_nan=False)
				return True
			except:
				return False

		if isinstance(metadata, dict):
			old_size = len(metadata)
			metadata = { k: v for k, v in metadata.items() if valid_json(v) }
			metadata = { k: v for k, v in metadata.items() if os.path.exists(os.path.join(path, k)) }
			new_size = len(metadata)
			if new_size != old_size:
				self._logger.info("Deleted {} stale or invalid entries from metadata for path {}".format(old_size - new_size,
				                                                                                         path))
				self._save_metadata(path, metadata)
			else:
				with self._get_metadata_lock(path):
					self._metadata_cache[path] = metadata
			return metadata
		else:
			return dict()

	def _save_metadata(self, path, metadata):
		import json

		with self._get_metadata_lock(path):
			self._metadata_cache[path] = metadata

		with self._get_persisted_metadata_lock(path):
			metadata_path = os.path.join(path, ".metadata.json")
			try:
				with atomic_write(metadata_path, mode='wb') as f:
					f.write(to_bytes(json.dumps(metadata, indent=4, separators=(",", ": "))))
			except Exception:
				self._logger.exception("Error while writing .metadata.json to {path}".format(**locals()))

	def _delete_metadata(self, path):
		with self._get_metadata_lock(path):
			if path in self._metadata_cache:
				del self._metadata_cache[path]

		with self._get_persisted_metadata_lock(path):
			metadata_files = (".metadata.json", ".metadata.yaml")
			for metadata_file in metadata_files:
				metadata_path = os.path.join(path, metadata_file)
				if os.path.exists(metadata_path):
					try:
						os.remove(metadata_path)
					except Exception:
						self._logger.exception("Error while deleting {metadata_file} from {path}".format(**locals()))

	@staticmethod
	def _copied_metadata(metadata, name):
		metadata = copy.copy(metadata)
		metadata[name] = copy.deepcopy(metadata.get(name, dict()))
		return metadata

	def _migrate_metadata(self, path):
		# we switched to json in 1.3.9 - if we still have yaml here, migrate it now
		import yaml
		import json

		with self._get_persisted_metadata_lock(path):
			metadata_path_yaml = os.path.join(path, ".metadata.yaml")
			metadata_path_json = os.path.join(path, ".metadata.json")

			if not os.path.exists(metadata_path_yaml):
				# nothing to migrate
				return

			if os.path.exists(metadata_path_json):
				# already migrated
				try:
					os.remove(metadata_path_yaml)
				except Exception:
					self._logger.exception("Error while removing .metadata.yaml from {path}".format(**locals()))
				return

			with io.open(metadata_path_yaml, 'rt', encoding='utf-8') as f:
				try:
					metadata = yaml.safe_load(f)
				except Exception:
					self._logger.exception("Error while reading .metadata.yaml from {path}".format(**locals()))
					return

			if not isinstance(metadata, dict):
				# looks invalid, ignore it
				return

			with atomic_write(metadata_path_json, mode='wb') as f:
				f.write(to_bytes(json.dumps(metadata, indent=4, separators=(",", ": "))))

			try:
				os.remove(metadata_path_yaml)
			except Exception:
				self._logger.exception("Error while removing .metadata.yaml from {path}".format(**locals()))

	@contextmanager
	def _get_metadata_lock(self, path):
		with self._metadata_lock_mutex:
			if path not in self._metadata_locks:
				import threading
				self._metadata_locks[path] = (0, threading.RLock())

			counter, lock = self._metadata_locks[path]
			counter += 1
			self._metadata_locks[path] = (counter, lock)

		yield lock

		with self._metadata_lock_mutex:
			counter = self._metadata_locks[path][0]
			counter -= 1
			if counter <= 0:
				del self._metadata_locks[path]
			else:
				self._metadata_locks[path] = (counter, lock)

	@contextmanager
	def _get_persisted_metadata_lock(self, path):
		with self._persisted_metadata_lock_mutex:
			if path not in self._persisted_metadata_locks:
				import threading
				self._persisted_metadata_locks[path] = (0, threading.RLock())

			counter, lock = self._persisted_metadata_locks[path]
			counter += 1
			self._persisted_metadata_locks[path] = (counter, lock)

		yield lock

		with self._persisted_metadata_lock_mutex:
			counter = self._persisted_metadata_locks[path][0]
			counter -= 1
			if counter <= 0:
				del self._persisted_metadata_locks[path]
			else:
				self._persisted_metadata_locks[path] = (counter, lock)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


import io
import sarge
import sys
import logging
import site
import threading
import os

import pkg_resources

from .commandline import CommandlineCaller, clean_ansi
from octoprint.util import to_unicode, to_native_str

_cache = dict(version=dict(), setup=dict())
_cache_mutex = threading.RLock()

class UnknownPip(Exception):
	pass

class PipCaller(CommandlineCaller):
	process_dependency_links = pkg_resources.Requirement.parse("pip>=1.5")
	no_cache_dir = pkg_resources.Requirement.parse("pip>=1.6")
	disable_pip_version_check = pkg_resources.Requirement.parse("pip>=6.0")
	no_use_wheel = pkg_resources.Requirement.parse("pip==1.5.0")
	broken = pkg_resources.Requirement.parse("pip>=6.0.1,<=6.0.3")

	@classmethod
	def clean_install_command(cls, args, pip_version, virtual_env, use_user, force_user):
		logger = logging.getLogger(__name__)
		args = list(args)

		# strip --process-dependency-links for versions that don't support it
		if not pip_version in cls.process_dependency_links and "--process-dependency-links" in args:
			logger.debug(
				"Found --process-dependency-links flag, version {} doesn't need that yet though, removing.".format(
					pip_version))
			args.remove("--process-dependency-links")

		# strip --no-cache-dir for versions that don't support it
		if not pip_version in cls.no_cache_dir and "--no-cache-dir" in args:
			logger.debug("Found --no-cache-dir flag, version {} doesn't support that yet though, removing.".format(pip_version))
			args.remove("--no-cache-dir")

		# strip --disable-pip-version-check for versions that don't support it
		if not pip_version in cls.disable_pip_version_check and "--disable-pip-version-check" in args:
			logger.debug("Found --disable-pip-version-check flag, version {} doesn't support that yet though, removing.".format(pip_version))
			args.remove("--disable-pip-version-check")

		# add --no-use-wheel for versions that otherwise break
		if pip_version in cls.no_use_wheel and not "--no-use-wheel" in args:
			logger.debug("Version {} needs --no-use-wheel to properly work.".format(pip_version))
			args.append("--no-use-wheel")

		# remove --user if it's present and a virtual env is detected
		if "--user" in args:
			if virtual_env or not site.ENABLE_USER_SITE:
				logger.debug("Virtual environment detected, removing --user flag.")
				args.remove("--user")
		# otherwise add it if necessary
		elif not virtual_env and site.ENABLE_USER_SITE and (use_user or force_user):
			logger.debug("pip needs --user flag for installations.")
			args.append("--user")

		return args

	def __init__(self, configured=None, ignore_cache=False, force_sudo=False,
	             force_user=False):
		CommandlineCaller.__init__(self)
		self._logger = logging.getLogger(__name__)

		self.configured = configured
		self.refresh = False
		self.ignore_cache = ignore_cache

		self.force_sudo = force_sudo
		self.force_user = force_user

		self._command = None
		self._version = None
		self._version_string = None
		self._use_sudo = False
		self._use_user = False
		self._virtual_env = False
		self._install_dir = None

		self.trigger_refresh()

		self.on_log_call = lambda *args, **kwargs: None
		self.on_log_stdout = lambda *args, **kwargs: None
		self.on_log_stderr = lambda *args, **kwargs: None

	def _reset(self):
		self._command = None
		self._version = None
		self._version_string = None
		self._use_sudo = False
		self._use_user = False
		self._install_dir = None

	def __le__(self, other):
		return self.version is not None and self.version <= other

	def __lt__(self, other):
		return self.version is not None and self.version < other

	def __ge__(self, other):
		return self.version is not None and self.version >= other

	def __gt__(self, other):
		return self.version is not None and self.version > other

	@property
	def command(self):
		return self._command

	@property
	def version(self):
		return self._version

	@property
	def version_string(self):
		return self._version_string

	@property
	def install_dir(self):
		return self._install_dir

	@property
	def use_sudo(self):
		return self._use_sudo

	@property
	def use_user(self):
		return self._use_user

	@property
	def virtual_env(self):
		return self._virtual_env

	@property
	def available(self):
		return self._command is not None

	def trigger_refresh(self):
		self._reset()
		try:
			self._setup_pip()
		except Exception:
			self._logger.exception("Error while discovering pip command")
			self._command = None
			self._version = None
		self.refresh = False

	def execute(self, *args, **kwargs):
		if self.refresh:
			self.trigger_refresh()

		if self._command is None:
			raise UnknownPip()

		arg_list = list(args)

		if "install" in arg_list:
			arg_list = self.clean_install_command(arg_list, self.version, self._virtual_env, self.use_user,
			                                      self.force_user)

		# add args to command
		if isinstance(self._command, list):
			command = self._command + list(arg_list)
		else:
			command = [self._command] + list(arg_list)

		# add sudo if necessary
		if self._use_sudo or self.force_sudo:
			command = ["sudo"] + command

		return self.call(command, **kwargs)

	def _setup_pip(self):
		pip_command, pip_sudo = self._get_pip_command()
		if pip_command is None:
			return

		# Determine the pip version

		self._logger.debug("Going to figure out pip's version")

		pip_version, version_segment = self._get_pip_version(pip_command)
		if pip_version is None:
			return

		if pip_version in self.__class__.broken:
			self._logger.error("This version of pip is known to have bugs that make it incompatible with how it needs "
			                   "to be used by OctoPrint. Please upgrade your pip version.")
			return

		# Now figure out if pip belongs to a virtual environment and if the
		# default installation directory is writable.
		#
		# The idea is the following: If OctoPrint is installed globally,
		# the site-packages folder is probably not writable by our user.
		# However, the user site-packages folder as usable by providing the
		# --user parameter during install is. This we may not use though if
		# the provided pip belongs to a virtual env (since that hiccups hard).
		#
		# So we figure out the installation directory, check if it's writable
		# and if not if pip belongs to a virtual environment. Only if the
		# installation directory is NOT writable by us but we also don't run
		# in a virtual environment may we proceed with the --user parameter.

		ok, pip_user, pip_virtual_env, pip_install_dir = self._check_pip_setup(pip_command)
		if not ok:
			if pip_install_dir:
				self._logger.error("Cannot use this pip install, can't write to the install dir and also can't use "
				                   "--user for installing. Check your setup and the permissions on {}.".format(pip_install_dir))
			else:
				self._logger.error("Cannot use this pip install, something's wrong with the python environment. "
				                   "Check the lines before.")
			return

		self._command = pip_command
		self._version = pip_version
		self._version_string = version_segment
		self._use_sudo = pip_sudo
		self._use_user = pip_user
		self._virtual_env = pip_virtual_env
		self._install_dir = pip_install_dir

	def _get_pip_command(self):
		pip_command = self.configured

		if pip_command is not None and pip_command.startswith("sudo "):
			pip_command = pip_command[len("sudo "):]
			pip_sudo = True
		else:
			pip_sudo = False

		if pip_command is None:
			pip_command = self.autodetect_pip()

		return pip_command, pip_sudo

	@classmethod
	def autodetect_pip(cls):
		commands = [[sys.executable, "-m", "pip"],
		            [os.path.join(os.path.dirname(sys.executable), "pip.exe" if sys.platform == "win32" else "pip")],

		            # this should be our last resort since it might fail thanks to using pip programmatically like
		            # that is not officially supported or sanctioned by the pip developers
		            [sys.executable, "-c", "import sys; sys.argv = ['pip'] + sys.argv[1:]; import pip; pip.main()"]]

		for command in commands:
			p = sarge.run(command + ["--version"], stdout=sarge.Capture(), stderr=sarge.Capture())
			if p.returncode == 0:
				logging.getLogger(__name__).info("Using \"{}\" as command to invoke pip".format(" ".join(command)))
				return command

		return None

	@classmethod
	def to_sarge_command(cls, pip_command, *args):
		if isinstance(pip_command, list):
			sarge_command = pip_command
		else:
			sarge_command = [pip_command]
		return sarge_command + list(args)

	def _get_pip_version(self, pip_command):
		# Debugging this with PyCharm/IntelliJ with Python plugin and no output is being
		# generated? PyCharm bug. Disable "Attach to subprocess automatically when debugging"
		# in IDE Settings or patch pydevd.py
		# -> https://youtrack.jetbrains.com/issue/PY-18365#comment=27-1290453

		pip_command_str = pip_command
		if isinstance(pip_command_str, list):
			pip_command_str = " ".join(pip_command_str)

		with _cache_mutex:
			if not self.ignore_cache and pip_command_str in _cache["version"]:
				self._logger.debug("Using cached pip version information for {}".format(pip_command_str))
				return _cache["version"][pip_command_str]

			sarge_command = self.to_sarge_command(pip_command, "--version")
			p = sarge.run(sarge_command, stdout=sarge.Capture(), stderr=sarge.Capture())

			if p.returncode != 0:
				self._logger.warning("Error while trying to run pip --version: {}".format(p.stderr.text))
				return None, None

			output = PipCaller._preprocess(p.stdout.text)
			# output should look something like this:
			#
			#     pip <version> from <path> (<python version>)
			#
			# we'll just split on whitespace and then try to use the second entry

			if not output.startswith("pip"):
				self._logger.warning("pip command returned unparseable output, can't determine version: {}".format(output))

			split_output = list(map(lambda x: x.strip(), output.split()))
			if len(split_output) < 2:
				self._logger.warning("pip command returned unparseable output, can't determine version: {}".format(output))

			version_segment = split_output[1]

			try:
				pip_version = pkg_resources.parse_version(version_segment)
			except Exception:
				self._logger.exception("Error while trying to parse version string from pip command")
				return None, None

			self._logger.info("Version of pip is {}".format(version_segment))

			result = pip_version, version_segment
			_cache["version"][pip_command_str] = result
			return result

	def _check_pip_setup(self, pip_command):
		pip_command_str = pip_command
		if isinstance(pip_command_str, list):
			pip_command_str = " ".join(pip_command_str)

		with _cache_mutex:
			if not self.ignore_cache and pip_command_str in _cache["setup"]:
				self._logger.debug("Using cached pip setup information for {}".format(pip_command_str))
				return _cache["setup"][pip_command_str]

			# This is horribly ugly and I'm sorry...
			#
			# While we can figure out the install directory, if that's writable and if a virtual environment
			# is active for pip that belongs to our sys.executable python instance by just checking some
			# variables, we can't for stuff like third party software we allow to update via the software
			# update plugin.
			#
			# What we do instead for these situations is try to install (and of course uninstall) the
			# testballoon dummy package, which collects that information for us. For pip <= 7 we could
			# have the testballoon provide us with the info needed through stdout, if pip was called
			# with --verbose anything printed to stdout within setup.py would be output. Pip 8 managed
			# to break this mechanism. Any (!) output within setup.py appears to be suppressed now, and
			# no combination of --log and multiple --verbose or -v arguments could get it to bring the
			# output back.
			#
			# So here's what we do now instead. Our sarge call sets an environment variable
			# "TESTBALLOON_OUTPUT" that points to a temporary file. If the testballoon sees that
			# variable set, it opens the file and writes to it the output it so far printed on stdout.
			# We then open the file and read in the data that way.
			#
			# Yeah, I'm not happy with that either. But as long as there's no way to otherwise figure
			# out for a generic pip command whether OctoPrint can even install anything with that
			# and if so how, well, that's how we'll have to do things.

			import os
			testballoon = os.path.join(os.path.realpath(os.path.dirname(__file__)), "piptestballoon")

			from octoprint.util import temppath
			with temppath() as testballoon_output_file:
				sarge_command = self.to_sarge_command(pip_command, "install", ".")
				try:
					# our testballoon is no real package, so this command will fail - that's ok though,
					# we only need the output produced within the pip environment
					sarge.run(sarge_command,
					          stdout=sarge.Capture(),
					          stderr=sarge.Capture(),
					          cwd=testballoon,
					          env=dict(TESTBALLOON_OUTPUT=to_native_str(testballoon_output_file)))
				except Exception:
					self._logger.exception("Error while trying to install testballoon to figure out pip setup")
					return False, False, False, None

				data = dict()
				with io.open(testballoon_output_file, 'rt', encoding='utf-8') as f:
					for line in f:
						key, value = line.split("=", 2)
						data[key] = value

			install_dir_str = data.get("PIP_INSTALL_DIR", None)
			virtual_env_str = data.get("PIP_VIRTUAL_ENV", None)
			writable_str = data.get("PIP_WRITABLE", None)

			if install_dir_str is not None and virtual_env_str is not None and writable_str is not None:
				install_dir = install_dir_str.strip()
				virtual_env = virtual_env_str.strip() == "True"
				writable = writable_str.strip() == "True"

				can_use_user_flag = not virtual_env and site.ENABLE_USER_SITE

				ok = writable or can_use_user_flag
				user_flag = not writable and can_use_user_flag

				self._logger.info("pip installs to {} (writable -> {}), --user flag needed -> {}, "
				                  "virtual env -> {}".format(install_dir,
				                                             "yes" if writable else "no",
				                                             "yes" if user_flag else "no",
				                                             "yes" if virtual_env else "no"))
				self._logger.info("==> pip ok -> {}".format("yes" if ok else "NO!"))

				# ok, enable user flag, virtual env yes/no, installation dir
				result = ok, user_flag, virtual_env, install_dir
				_cache["setup"][pip_command_str] = result
				return result
			else:
				self._logger.error("Could not detect desired output from testballoon install, got this instead: {!r}".format(data))
				return False, False, False, None

	def _preprocess_lines(self, *lines):
		return list(map(self._preprocess, lines))

	@staticmethod
	def _preprocess(text):
		"""
		Strips ANSI and VT100 cursor control characters from line and makes sure it's a unicode.

		Parameters:
		    text (str or unicode): The text to process

		Returns:
		    (unicode) The processed text as a unicode, stripped of ANSI and VT100 cursor show/hide codes

		Example::

		    >>> text = b'some text with some\x1b[?25h ANSI codes for \x1b[31mred words\x1b[39m and\x1b[?25l also some cursor control codes'
		    >>> PipCaller._preprocess(text) # doctest: +ALLOW_UNICODE
		    'some text with some ANSI codes for red words and also some cursor control codes'
		"""
		return to_unicode(clean_ansi(text))

class LocalPipCaller(PipCaller):
	"""
	The LocalPipCaller always uses the pip instance associated with
	sys.executable.
	"""

	def _get_pip_command(self):
		return self.autodetect_pip(), False

	def _check_pip_setup(self, pip_command):
		import sys
		import os
		from distutils.sysconfig import get_python_lib

		virtual_env = hasattr(sys, "real_prefix")
		install_dir = get_python_lib()
		writable = os.access(install_dir, os.W_OK)

		can_use_user_flag = not virtual_env and site.ENABLE_USER_SITE
		user_flag = not writable and can_use_user_flag

		ok = writable or can_use_user_flag

		self._logger.info("pip installs to {} (writable -> {}), --user flag needed -> {}, "
		                  "virtual env -> {}".format(install_dir,
		                                             "yes" if writable else "no",
		                                             "yes" if user_flag else "no",
		                                             "yes" if virtual_env else "no"))
		self._logger.info("==> pip ok -> {}".format("yes" if ok else "NO!"))

		return ok, \
		       user_flag, \
		       virtual_env, \
		       install_dir
